import os
import logging
from typing import Dict, List, Tuple, Union
from collections import OrderedDict
from tempfile import NamedTemporaryFile

import pandas as pd
import pygsheets as ps
import json

from schematic.schemas.generator import SchemaGenerator
from schematic.utils.google_api_utils import (
    build_credentials,
    execute_google_api_requests,
    build_service_account_creds,
)
from schematic.utils.df_utils import update_df, load_df
from schematic.utils.validate_utils import rule_in_rule_list

#TODO: This module should only be aware of the store interface
# we shouldn't need to expose Synapse functionality explicitly
from schematic.store.synapse import SynapseStorage

from schematic import CONFIG
from schematic.utils.google_api_utils import export_manifest_drive_service
from openpyxl import load_workbook
from pathlib import Path

logger = logging.getLogger(__name__)



class ManifestGenerator(object):
    def __init__(
        self,
        path_to_json_ld: str,  # JSON-LD file to be used for generating the manifest
        alphabetize_valid_values: str = 'ascending',
        title: str = None,  # manifest sheet title
        root: str = None,
        additional_metadata: Dict = None,
        oauth: bool = True,
        use_annotations: bool = False,
    ) -> None:

        if oauth:
            # if user wants to use OAuth for Google authentication
            # use credentials.json and create token.pickle file
            services_creds = build_credentials()
        else:
            # if not oauth then use service account credentials
            services_creds = build_service_account_creds()

        # google service for Sheet API
        self.sheet_service = services_creds["sheet_service"]

        # google service for Drive API
        self.drive_service = services_creds["drive_service"]

        # google service credentials object
        self.creds = services_creds["creds"]

        # schema root
        self.root = root

        # alphabetize valid values
        self.alphabetize = alphabetize_valid_values

        # manifest title
        self.title = title
        if self.title is None:
            self.title = f"{self.root} - Manifest"

        # Whether to use existing annotations during manifest generation
        self.use_annotations = use_annotations

        # Warn about limited feature support for `use_annotations`
        if self.use_annotations:
            logger.warning(
                "The `use_annotations` option is currently only supported "
                "when there is no manifest file for the dataset in question."
            )

        # SchemaGenerator() object
        self.sg = SchemaGenerator(path_to_json_ld)

        # additional metadata to add to manifest
        self.additional_metadata = additional_metadata

        # Determine whether current data type is file-based
        is_file_based = False
        if self.root:
            is_file_based = "Filename" in self.sg.get_node_dependencies(self.root)
        self.is_file_based = is_file_based

    def _attribute_to_letter(self, attribute, manifest_fields):
        """Map attribute to column letter in a google sheet"""

        # find index of attribute in manifest field
        column_idx = manifest_fields.index(attribute)

        # return the google sheet letter representation of the column index
        return self._column_to_letter(column_idx)

    def _column_to_letter(self, column):
        """Find google sheet letter representation of a column index integer"""
        character = chr(ord("A") + column % 26)
        remainder = column // 26
        if column >= 26:
            return self._column_to_letter(remainder - 1) + character
        else:
            return character

    def _columns_to_sheet_ranges(self, column_idxs):
        """map a set of column indexes to a set of Google sheet API ranges: each range includes exactly one column"""
        ranges = []

        for column_idx in column_idxs:
            col_range = {
                "startColumnIndex": column_idx,
                "endColumnIndex": column_idx + 1,
            }

            ranges.append(col_range)

        return ranges

    def _column_to_cond_format_eq_rule(
        self, column_idx: int, condition_argument: str, required: bool = False
    ) -> dict:
        """Given a column index and an equality argument (e.g. one of valid values for the given column fields), generate a conditional formatting rule based on a custom formula encoding the logic:

        'if a cell in column idx is equal to condition argument, then set specified formatting'
        """

        col_letter = self._column_to_letter(column_idx)

        if not required:
            bg_color = CONFIG["style"]["google_manifest"].get(
                "opt_bg_color", {"red": 1.0, "green": 1.0, "blue": 0.9019,},
            )
        else:
            bg_color = CONFIG["style"]["google_manifest"].get(
                "req_bg_color", {"red": 0.9215, "green": 0.9725, "blue": 0.9803,},
            )

        boolean_rule = {
            "condition": {
                "type": "CUSTOM_FORMULA",
                "values": [
                    {
                        "userEnteredValue": "=$"
                        + col_letter
                        + '1 = "'
                        + condition_argument
                        + '"'
                    }
                ],
            },
            "format": {"backgroundColor": bg_color},
        }

        return boolean_rule

    def _gdrive_copy_file(self, origin_file_id, copy_title):
        """Copy an existing file.

        Args:
            origin_file_id: ID of the origin file to copy.
            copy_title: Title of the copy.

        Returns:
            The copied file if successful, None otherwise.
        """
        copied_file = {"name": copy_title}

        # return new copy sheet ID
        return (
            self.drive_service.files()
            .copy(fileId=origin_file_id, body=copied_file)
            .execute()["id"]
        )

    def _create_empty_manifest_spreadsheet(self, title):
        if CONFIG["style"]["google_manifest"]["master_template_id"]:

            # if provided with a template manifest google sheet, use it
            spreadsheet_id = self._gdrive_copy_file(
                CONFIG["style"]["google_manifest"]["master_template_id"], title
            )

        else:
            # if no template, create an empty spreadsheet
            spreadsheet = (
                self.sheet_service.spreadsheets()
                .create(body=spreadsheet, fields="spreadsheetId")
                .execute()
            )
            spreadsheet_id = spreadsheet.get("spreadsheetId")

        return spreadsheet_id

    def _get_cell_borders(self, cell_range):

        # set border style request
        color = {
            "red": 226.0 / 255.0,
            "green": 227.0 / 255.0,
            "blue": 227.0 / 255.0,
        }

        border_style_req = {
            "updateBorders": {
                "range": cell_range,
                "top": {"style": "SOLID", "width": 2, "color": color},
                "bottom": {"style": "SOLID", "width": 2, "color": color},
                "left": {"style": "SOLID", "width": 2, "color": color},
                "right": {"style": "SOLID", "width": 2, "color": color},
                "innerHorizontal": {"style": "SOLID", "width": 2, "color": color},
                "innerVertical": {"style": "SOLID", "width": 2, "color": color},
            }
        }

        return border_style_req

    def _set_permissions(self, fileId):
        def callback(request_id, response, exception):
            if exception:
                # Handle error
                logger.error(exception)
            else:
                logger.info(f"Permission Id: {response.get('id')}")

        batch = self.drive_service.new_batch_http_request(callback=callback)

        worldPermission = {"type": "anyone", "role": "writer"}

        batch.add(
            self.drive_service.permissions().create(
                fileId=fileId, body=worldPermission, fields="id",
            )
        )
        batch.execute()

    def _get_column_data_validation_values(
        self,
        spreadsheet_id,
        valid_values,
        column_id,
        strict,
        validation_type="ONE_OF_LIST",
        custom_ui=True,
        input_message="Choose one from dropdown",
    ):

        # set validation strictness to config file default if None indicated.
        if strict == None:
            strict = CONFIG["style"]["google_manifest"].get("strict_validation", True)

        # get valid values w/o google sheet header
        values = [valid_value["userEnteredValue"] for valid_value in valid_values]
        
        if self.alphabetize and self.alphabetize.lower().startswith('a'):
            values.sort(reverse=False, key=str.lower)
        elif self.alphabetize and self.alphabetize.lower().startswith('d'):
            values.sort(reverse=True, key=str.lower)
        

        if validation_type == "ONE_OF_RANGE":

            # store valid values explicitly in workbook at the provided range to use as validation values
            target_col_letter = self._column_to_letter(column_id)
            body = {"majorDimension": "COLUMNS", "values": [values]}
            target_range = (
                "Sheet2!"
                + target_col_letter
                + "2:"
                + target_col_letter
                + str(len(values) + 1)
            )
            valid_values = [{"userEnteredValue": "=" + target_range}]

            response = (
                self.sheet_service.spreadsheets()
                .values()
                .update(
                    spreadsheetId=spreadsheet_id,
                    range=target_range,
                    valueInputOption="RAW",
                    body=body,
                )
                .execute()
            )

        # setup validation data request body
        validation_body = {
            "requests": [
                {
                    "setDataValidation": {
                        "range": {
                            "startRowIndex": 1,
                            "startColumnIndex": column_id,
                            "endColumnIndex": column_id + 1,
                        },
                        "rule": {
                            "condition": {
                                "type": validation_type,
                                "values": valid_values,
                            },
                            "inputMessage": input_message,
                            "strict": strict,
                            "showCustomUi": custom_ui,
                        },
                    }
                }
            ]
        }

        return validation_body

    def _get_valid_values_from_jsonschema_property(self, prop: dict) -> List[str]:
        """Get valid values for a manifest attribute based on the corresponding
        values of node's properties in JSONSchema

        Args:
            prop: node properties - jsonschema dictionary

        Returns:
            List of valid values
        """

        if "enum" in prop:
            return prop["enum"]
        elif "items" in prop:
            return prop["items"]["enum"]
        else:
            return []


    def _get_json_schema(self, json_schema_filepath: str) -> Dict:
        """Open json schema as a dictionary.
        Args:
            json_schema_filepath(str): path to json schema file
        Returns:
            Dictionary, containing portions of the json schema
        """
        if not json_schema_filepath:
            # if no json schema is provided; there must be
            # schema explorer defined for schema.org schema
            # o.w. this will throw an error
            # TODO: catch error
            json_schema = self.sg.get_json_schema_requirements(self.root, self.title)
        else:
            with open(json_schema_filepath) as jsonfile:
                json_schema = json.load(jsonfile)
        return json_schema

    def _get_required_metadata_fields(self, json_schema, fields):
        """For the root node gather dependency requirements (all attributes linked to this node)
        and corresponding allowed values constraints (i.e. valid values).

        Args:
            json_schema(dict): representing a handful of values
                representing the data model, including: '$schema', '$id', 'title',
                'type', 'properties', 'required'
            fields(list[str]): fields/attributes to search
        Returns:
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
        """
        required_metadata_fields = {}
        for req in fields:
            required_metadata_fields[
                req
            ] = self._get_valid_values_from_jsonschema_property(
                json_schema["properties"][req]
            )
            # the following line may not be needed
            json_schema["properties"][req]["enum"] = required_metadata_fields[req]
        return required_metadata_fields

    def _gather_dependency_requirements(self, json_schema, required_metadata_fields):
        """Gathering dependency requirements and allowed value constraints
            for conditional dependencies, if any
        TODO:
            Refactor
        Args:
            json_schema(dict): representing a handful of values
                representing the data model, including: '$schema', '$id', 'title',
                'type', 'properties', 'required'
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
        Returns:
            required_metadata_fields(dict):
                updates dictionary to additional attributes, if applicable,
                as keys with their corresponding valid values.
        """

        if "allOf" in json_schema:
            for conditional_reqs in json_schema["allOf"]:
                if "required" in conditional_reqs["if"]:
                    for req in conditional_reqs["if"]["required"]:
                        if req in conditional_reqs["if"]["properties"]:
                            if not req in required_metadata_fields:
                                if req in json_schema["properties"]:
                                    required_metadata_fields[
                                        req
                                    ] = self._get_valid_values_from_jsonschema_property(
                                        json_schema["properties"][req]
                                    )
                                else:
                                    required_metadata_fields[
                                        req
                                    ] = self._get_valid_values_from_jsonschema_property(
                                        conditional_reqs["if"]["properties"][req]
                                    )

                    for req in conditional_reqs["then"]["required"]:
                        if not req in required_metadata_fields:
                            if req in json_schema["properties"]:
                                required_metadata_fields[
                                    req
                                ] = self._get_valid_values_from_jsonschema_property(
                                    json_schema["properties"][req]
                                )

        return required_metadata_fields

    def _add_root_to_component(self, required_metadata_fields):
        """If 'Component' is in the column set, add root node as a
        metadata component entry in the first row of that column.
        Args:
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
        Return:
            updates self.additional_metadata if appropriate to
            contain {'Component': [self.root]}
        """
        if "Component" in required_metadata_fields.keys():
            # check if additional metadata has actually been instantiated in the
            # constructor (it's optional) if not, instantiate it
            if not self.additional_metadata:
                self.additional_metadata = {}
            if self.is_file_based and 'Filename' in self.additional_metadata:
                self.additional_metadata["Component"] = [self.root] * max(
                    1, len(self.additional_metadata["Filename"])
                )
            else:
                self.additional_metadata["Component"] = [self.root]

        return

    def _get_additional_metadata(self, required_metadata_fields: dict) -> dict:
        """Add additional metadata as entries to columns.
        Args:
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
        Return:
            required_metadata_fields(dict): Updated required_metadata_fields to include additional metadata
                that is to be added to specified columns
        """
        self._add_root_to_component(required_metadata_fields)
        # if additional metadata is provided append columns (if those do not exist already)
        if self.additional_metadata:
            for column in self.additional_metadata.keys():
                if not column in required_metadata_fields:
                    required_metadata_fields[column] = []
        return required_metadata_fields

    def _get_column_range_and_order(self, required_metadata_fields):
        """Find the alphabetical range of columns and sort them.
        Args:
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
        Returns:
            end_col_letter (chr): google sheet letter representation of a column index integer
            ordered_metadata_fields(List[list]): Contining all the attributes to
            add as columns, ordered
        """
        # determining columns range
        end_col = len(required_metadata_fields.keys())
        end_col_letter = self._column_to_letter(end_col)

        # order columns header (since they are generated based on a json schema, which is a dict)
        ordered_metadata_fields = [list(required_metadata_fields.keys())]
        ordered_metadata_fields[0] = self.sort_manifest_fields(
            ordered_metadata_fields[0]
        )
        return end_col_letter, ordered_metadata_fields

    def _gs_add_and_format_columns(self, required_metadata_fields, spreadsheet_id):
        """Add columns to the google sheet and format them.
        Args:
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
            spreadsheet_id(str): id of the google sheet
        Returns:
            respones: In return for a request, used by google sheet api to add columns
                to the sheet being generated.
            ordered_metadata_fields(List[list]): contining all the attributes to
            add as columns, ordered
        """

        end_col_letter, ordered_metadata_fields = self._get_column_range_and_order(
            required_metadata_fields
        )

        body = {"values": ordered_metadata_fields}

        range = "Sheet1!A1:" + str(end_col_letter) + "1"

        # adding columns
        self.sheet_service.spreadsheets().values().update(
            spreadsheetId=spreadsheet_id, range=range, valueInputOption="RAW", body=body
        ).execute()

        # adding columns to 2nd sheet that can be used for storing data validation ranges (this avoids limitations on number of dropdown items in excel and openoffice)
        range = "Sheet2!A1:" + str(end_col_letter) + "1"
        self.sheet_service.spreadsheets().values().update(
            spreadsheetId=spreadsheet_id, range=range, valueInputOption="RAW", body=body
        ).execute()

        # format column header row
        header_format_body = {
            "requests": [
                {
                    "repeatCell": {
                        "range": {"startRowIndex": 0, "endRowIndex": 1},
                        "cell": {
                            "userEnteredFormat": {
                                "backgroundColor": {
                                    "red": 224.0 / 255,
                                    "green": 224.0 / 255,
                                    "blue": 224.0 / 255,
                                },
                                "horizontalAlignment": "CENTER",
                                "textFormat": {
                                    "foregroundColor": {
                                        "red": 0.0 / 255,
                                        "green": 0.0 / 255,
                                        "blue": 0.0 / 255,
                                    },
                                    "fontSize": 8,
                                    "bold": True,
                                },
                            }
                        },
                        "fields": "userEnteredFormat(backgroundColor,textFormat,horizontalAlignment)",
                    }
                },
                {
                    "updateSheetProperties": {
                        "properties": {"gridProperties": {"frozenRowCount": 1}},
                        "fields": "gridProperties.frozenRowCount",
                    }
                },
                {
                    "autoResizeDimensions": {
                        "dimensions": {"dimension": "COLUMNS", "startIndex": 0}
                    }
                },
            ]
        }

        response = (
            self.sheet_service.spreadsheets()
            .batchUpdate(spreadsheetId=spreadsheet_id, body=header_format_body)
            .execute()
        )
        return response, ordered_metadata_fields

    def _gs_add_additional_metadata(
        self, required_metadata_fields, ordered_metadata_fields, spreadsheet_id
    ):
        """Adding additional metadata values, if needed. Add value-constraints
        from data model as dropdowns. Batch google API request to
        create metadata template.
        Args:
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
            ordered_metadata_fields(List[list]): contining all the attributes to
            add as columns, ordered
            spreadsheet_id(str): id for the google sheet
        Returns: Response(dict): For batch update to create metadata
        template.
        """
        data = []

        for i, req in enumerate(ordered_metadata_fields[0]):
            values = required_metadata_fields[req]

            if self.additional_metadata and req in self.additional_metadata:
                values = self.additional_metadata[req]
                target_col_letter = self._column_to_letter(i)

                range_vals = (
                    target_col_letter + "2:" + target_col_letter + str(len(values) + 1)
                )

                data.append(
                    {
                        "range": range_vals,
                        "majorDimension": "COLUMNS",
                        "values": [values],
                    }
                )

        batch_update_values_request_body = {
            # How the input data should be interpreted.
            "valueInputOption": "RAW",
            # The new values to apply to the spreadsheet.
            "data": data,
        }

        response = (
            self.sheet_service.spreadsheets()
            .values()
            .batchUpdate(
                spreadsheetId=spreadsheet_id, body=batch_update_values_request_body
            )
            .execute()
        )
        return response

    def _request_update_base_color(self, i: int, color={"red": 1.0}):
        """
        Change color of text in column we are validating
        to red.
        """
        vr_format_body = {
            "requests": [
                {
                    "repeatCell": {
                        "range": {
                            "startColumnIndex": i,
                            "endColumnIndex": i + 1,
                            "startRowIndex": 1,
                        },
                        "cell": {
                            "userEnteredFormat": {
                                "textFormat": {"foregroundColor": color}
                            }
                        },
                        "fields": "userEnteredFormat(textFormat)",
                    }
                }
            ]
        }
        return vr_format_body

    def _request_regex_vr(self, gs_formula, i:int, text_color={"red": 1}):
        """
        Generate request to change font color to black upon corretly formatted
        user entry.
        """
        requests_vr = {
            "requests": [
                {
                    "addConditionalFormatRule": {
                        "rule": {
                            "ranges": {
                                "startColumnIndex": i,
                                "endColumnIndex": i + 1,
                                "startRowIndex": 1,
                            },
                            "booleanRule": {
                                "condition": {
                                    "type": "CUSTOM_FORMULA",
                                    "values": gs_formula,
                                },
                                "format": {
                                    "textFormat": {
                                        "foregroundColor": text_color
                                    }
                                },
                            },
                        },
                        "index": 0,
                    }
                }
            ]
        }
        return requests_vr

    def _request_regex_match_vr_formatting(self, validation_rules: List[str], i: int,
        spreadsheet_id: str, requests_body: dict,
        ):
        """
        Purpose:
            - Apply regular expression validaiton rules to google sheets.
            - Will only be run if the regex module specified in the validation
            rules is 'match'.
                - This is because of the limitations of google sheets regex.
                - Users can additionally add 'strict' to the end of their validation
                rules. This would allow the rule itself to set the level of strictness,
                otherwise it would fall to the default level set in the config file.
            - Will do the following:
                - In google sheets user entry text will initially appear red.
                - Upon correct format entry, text will turn black.
                - If incorrect format is entered a validation error will pop up.
        Input:
            validation_rules: List[str], defines the validation rules
                applied to a particular column.
            i: int, defines current column.
            requests_body: dict, containing all the update requests to add to the gs
        Returns:
            requests_body: updated with additional formatting requests if regex match
                is specified.
        """
        split_rules = validation_rules[0].split(" ")
        if split_rules[0] == "regex" and split_rules[1] == "match":
            # Set things up:
            ## Extract the regular expression we are validating against.
            regular_expression = split_rules[2]
            ## Define text color to update to upon correct user entry
            text_color = {"red": 0, "green": 0, "blue": 0}
            ## Define google sheets regular expression formula
            gs_formula = [
                {
                    "userEnteredValue": '=REGEXMATCH(INDIRECT("RC",FALSE), "{}")'.format(
                        regular_expression
                    )
                }
            ]
            ## Set validaiton strictness based on user specifications.
            strict = None
            if split_rules[-1].lower() == "strict":
                strict = True

            ## Create error message for users if they enter value with incorrect formatting
            input_message = (
                f"Values in this column are being validated "
                f"against the following regular expression ({regular_expression}) "
                f"to ensure for accuracy. Please re-enter value according to these "
                f"formatting rules"
            )

            # Create Requests:
            ## Change request to change the text color of the column we are validating to red.
            requests_vr_format_body = self._request_update_base_color(
                i,
                color={
                    "red": 232.0 / 255.0,
                    "green": 80.0 / 255.0,
                    "blue": 70.0 / 255.0,
                }
            )

            ## Create request to for conditionally formatting user input.
            requests_vr = self._request_regex_vr(gs_formula, i, text_color)

            ## Create request to generate data validator.
            requests_data_validation_vr = self._get_column_data_validation_values(
                spreadsheet_id,
                valid_values=gs_formula,
                column_id=i,
                strict=strict,
                custom_ui=False,
                input_message=input_message,
                validation_type="CUSTOM_FORMULA",
            )

            requests_body["requests"].append(
                requests_vr_format_body["requests"]
            )
            requests_body["requests"].append(requests_vr["requests"])
            requests_body["requests"].append(
                requests_data_validation_vr["requests"]
            )
        return requests_body


    def _request_row_format(self, i, req):
        """Adding description to headers, this is not executed if
        only JSON schema is defined. Also formatting required columns.
        Args:
            i (int): column index
            req (str): column name
        Returns:
            notes_body["requests"] (dict): with information on note
                to add to the column header. This notes body will be added to a request.
        """
        if self.sg.se:
            # get node definition
            note = self.sg.get_node_definition(req)

            notes_body = {
                "requests": [
                    {
                        "updateCells": {
                            "range": {
                                "startRowIndex": 0,
                                "endRowIndex": 1,
                                "startColumnIndex": i,
                                "endColumnIndex": i + 1,
                            },
                            "rows": [{"values": [{"note": note}]}],
                            "fields": "note",
                        }
                    }
                ]
            }

            return notes_body["requests"]
        else:
            return

    def _request_note_valid_values(self, i, req, validation_rules, valid_values):
        """If the node validation rule is 'list' and the node also contains
        valid values, add a note a note with instructions on adding a list of
        multiple values with the multi-select option in google sheets

        TODO: add validation and QC rules "compiler/generator" class elsewhere
            for now have the list logic here
        Args:
            i (int): column index
            req (str): column name
            validation_rules(list[str]): containing all relevant
                rules applied to node/column
            valid_values(list[str]): containing all valid values defined in the
                data model for this node/column

        Returns:
            notes_body["requests"] (dict): with information on note
                to add to the column header, about using multiselect.
                This notes body will be added to a request.
        """            
       
        if rule_in_rule_list("list", validation_rules) and valid_values:
            note = "From 'Selection options' menu above, go to 'Select multiple values', check all items that apply, and click 'Save selected values'"
            notes_body = {
                "requests": [
                    {
                        "repeatCell": {
                            "range": {
                                "startRowIndex": 1,
                                "startColumnIndex": i,
                                "endColumnIndex": i + 1,
                            },
                            "cell": {"note": note},
                            "fields": "note",
                        }
                    }
                ]
            }
            return notes_body["requests"]
        elif rule_in_rule_list("list", validation_rules) and not valid_values:
            note = "Please enter values as a comma separated list. For example: XX, YY, ZZ"
            notes_body = {
                "requests": [
                    {
                        "repeatCell": {
                            "range": {
                                "startRowIndex": 1,
                                "startColumnIndex": i,
                                "endColumnIndex": i + 1,
                            },
                            "cell": {"note": note},
                            "fields": "note",
                        }
                    }
                ]
            }
            return notes_body["requests"]
        else:
            return

    def _request_notes_comments(self, i, req, json_schema):
        """Update background colors so that columns that are required are highlighted
        Args:
            i (int): column index
            req (str): column name
            json_schema(dict): representing a handful of values
                representing the data model, including: '$schema', '$id', 'title',
                'type', 'properties', 'required'
        Returns:
            req_format_body["requests"] (dict): specifing the format updating for
                required attributes/columns
        """
        # check if attribute is required and set a corresponding color
        if req in json_schema["required"]:
            bg_color = CONFIG["style"]["google_manifest"].get(
                "req_bg_color",
                {
                    "red": 0.9215,
                    "green": 0.9725,
                    "blue": 0.9803,
                },
            )

            req_format_body = {
                "requests": [
                    {
                        "repeatCell": {
                            "range": {
                                "startColumnIndex": i,
                                "endColumnIndex": i + 1,
                            },
                            "cell": {
                                "userEnteredFormat": {"backgroundColor": bg_color}
                            },
                            "fields": "userEnteredFormat(backgroundColor)",
                        }
                    }
                ]
            }
            return req_format_body["requests"]
        else:
            return

    def _request_cell_borders(self):
        """Get cell border color specifications.
        Args: None
        Returns:
            Dict, containing cell border design specifications.
        """
        # setting cell borders
        cell_range = {
            "sheetId": 0,
            "startRowIndex": 0,
        }
        return self._get_cell_borders(cell_range)

    def _request_dropdown_or_multi(
        self, i, req_vals, spreadsheet_id, validation_rules, valid_values
    ):
        """Generating sheet api request to populate a dropdown or a multi selection UI
        Args:
            i (int): column index
            req_vals (list[dict]): dict for each valid value
                key: 'userEnteredValue'
                value: str, valid value
            spreadsheet_id (str): id for the google sheet
            validation_rules(list[str]): containing all relevant
                rules applied to node/column
            valid_values(list[str]): containing all valid values defined in the
                data model for this node/column
        Returns:
            validation_body: dict
        """
        if len(req_vals) > 0 and not rule_in_rule_list("list", validation_rules):
            # if more than 0 values in dropdown use ONE_OF_RANGE type of validation
            # since excel and openoffice
            # do not support other kinds of data validation for
            # larger number of items (even if individual items are not that many
            # excel has a total number of characters limit per dropdown...)
            validation_body = self._get_column_data_validation_values(
                spreadsheet_id, req_vals, i, strict=None, validation_type="ONE_OF_RANGE"
            )
        elif rule_in_rule_list("list", validation_rules) and valid_values:
            # if list is in validation rule attempt to create a multi-value
            # selection UI, which requires explicit valid values range in
            # the spreadsheet
            # set "strict" parameter to false to allow users enter multiple values on google sheet
            validation_body = self._get_column_data_validation_values(
                spreadsheet_id,
                req_vals,
                i,
                strict=False,
                custom_ui=False,
                input_message="",
                validation_type="ONE_OF_RANGE",
            )
        else:
            validation_body = self._get_column_data_validation_values(
                spreadsheet_id, req_vals, i, strict=None
            )
        return validation_body["requests"]

    def _dependency_formatting(
        self, i, req_val, ordered_metadata_fields, val_dependencies,
        dependency_formatting_body
    ):
        """If there are additional attribute dependencies find the corresponding
        fields that need to be filled in and construct conditional formatting rules
        indicating the dependencies need to be filled in.

        set target ranges for this rule
        i.e. dependency attribute columns that will be formatted

        Args:
            i (int): column index
            req_val (str): node name
            ordered_metadata_fields(List[list]): contining all the attributes to
            add as columns, ordered
            val_depenencies (list[str]): dependencies
        Returns:
            dependency_formatting_body["requests"] (list):
                specifies gs conditional formatting per val_dependency
        """

        # find dependency column indexes
        # note that dependencies values must be in index
        # TODO: catch value error that shouldn't happen
        column_idxs = [
            ordered_metadata_fields[0].index(val_dep) for val_dep in val_dependencies
        ]

        # construct ranges based on dependency column indexes
        rule_ranges = self._columns_to_sheet_ranges(column_idxs)
        # go over valid value dependencies
        dependency_formatting_body = {"requests": []}
        for j, val_dep in enumerate(val_dependencies):
            is_required = False

            if self.sg.is_node_required(val_dep):
                is_required = True
            else:
                is_required = False

            # construct formatting rule
            formatting_rule = self._column_to_cond_format_eq_rule(
                i, req_val, required=is_required
            )

            # construct conditional format rule
            conditional_format_rule = {
                "addConditionalFormatRule": {
                    "rule": {
                        "ranges": rule_ranges[j],
                        "booleanRule": formatting_rule,
                    },
                    "index": 0,
                }
            }
            dependency_formatting_body["requests"].append(
                            conditional_format_rule
                        )
        return dependency_formatting_body["requests"]

    def _request_dependency_formatting(
        self, i, req_vals, ordered_metadata_fields, requests_body
    ):
        """Gather all the formattng for node dependencies.
        Args:
            i (int): column index
            req_val (str): node name
            ordered_metadata_fields(List[list]): contining all the attributes to
            add as columns, ordered
            requests_body(dict):
                containing all the update requests to add to the gs
        Return:
            requests_body (dict): adding the conditional
            formatting rules to apply
        """
        for req_val in req_vals:
            # get this required/valid value's node label in schema, based on display name (i.e. shown to the user in a dropdown to fill in)
            req_val = req_val["userEnteredValue"]
            req_val_node_label = self.sg.get_node_label(req_val)
            if not req_val_node_label:
                # if this node is not in the graph
                # continue - there are no dependencies for it
                continue
            # check if this required/valid value has additional dependency attributes
            val_dependencies = self.sg.get_node_dependencies(
                req_val_node_label, schema_ordered=False
            )

            # prepare request calls
            dependency_formatting_body = {"requests": []}

            # set conditiaon formatting for dependencies.
            if val_dependencies:
                dependency_formatting_body["requests"] = self._dependency_formatting(
                    i, req_val, ordered_metadata_fields, val_dependencies,
                    dependency_formatting_body
                )

            if dependency_formatting_body["requests"]:
                requests_body["requests"].append(
                    dependency_formatting_body["requests"]
                )
        return requests_body

    def _create_requests_body(
        self,
        required_metadata_fields,
        ordered_metadata_fields,
        json_schema,
        spreadsheet_id,
    ):
        """Create and store all formatting changes for the google sheet to
        execute at once.
        Args:
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
            ordered_metadata_fields: List[list], contining all the attributes to
                add as columns, ordered
            json_schema: dict, representing a handful of values
                representing the data model, including: '$schema', '$id', 'title',
                'type', 'properties', 'required'
            spreadsheet_id: str, of the id for the google sheet
        Return:
            requests_body(dict):
                containing all the update requests to add to the gs
        """
        # store all requests to execute at once
        requests_body = {}
        requests_body["requests"] = []
        for i, req in enumerate(ordered_metadata_fields[0]):
            # Gather validation rules and valid values for attribute
            validation_rules = self.sg.get_node_validation_rules(req)

            if validation_rules:
                requests_body =self._request_regex_match_vr_formatting(
                        validation_rules, i, spreadsheet_id, requests_body
                        )

            if req in json_schema["properties"].keys():
                valid_values = self._get_valid_values_from_jsonschema_property(
                    json_schema["properties"][req]
                )
            else:
                valid_values = []

            # Set row formatting
            get_row_formatting = self._request_row_format(i, req)
            if get_row_formatting:
                requests_body["requests"].append(get_row_formatting)

            # Add notes to headers to provide descriptions of the attribute
            header_notes = self._request_notes_comments(i, req, json_schema)
            if header_notes:
                requests_body["requests"].append(header_notes)
            # Add note on how to use multi-select, when appropriate
            note_vv = self._request_note_valid_values(
                i, req, validation_rules, valid_values
            )
            if note_vv:
                requests_body["requests"].append(note_vv)

            # Adding value-constraints, if any
            values = required_metadata_fields[req]
            req_vals = [{"userEnteredValue": value} for value in values if value]
            if not req_vals:
                continue

            # Create dropdown or multi-select options, where called for
            create_dropdown = self._request_dropdown_or_multi(
                i, req_vals, spreadsheet_id, validation_rules, valid_values
            )
            if create_dropdown:
                requests_body["requests"].append(create_dropdown)

            # generate a conditional format rule for each required value (i.e. valid value)
            # for this field (i.e. if this field is set to a valid value that may require additional
            # fields to be filled in, these additional fields will be formatted in a custom style (e.g. red background)

            requests_body = self._request_dependency_formatting(i, req_vals, ordered_metadata_fields, requests_body)
           
        # Set borders formatting
        borders_formatting = self._request_cell_borders()
        if borders_formatting:
            requests_body["requests"].append(borders_formatting)
        return requests_body

    def _create_empty_gs(self, required_metadata_fields, json_schema, spreadsheet_id):
        """Generate requests to add columns and format the google sheet.
        Args:
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
            json_schema: dict, representing a handful of values
                representing the data model, including: '$schema', '$id', 'title',
                'type', 'properties', 'required'
            spreadsheet_id: str, of the id for the google sheet
        Returns:
            manifest_url (str): url of the google sheet manifest.
        """
        # adding columns to manifest sheet
        response, ordered_metadata_fields = self._gs_add_and_format_columns(
            required_metadata_fields, spreadsheet_id
        )

        # Add additional metadata
        response = self._gs_add_additional_metadata(
            required_metadata_fields, ordered_metadata_fields, spreadsheet_id
        )

        # Create requests body to add additional formatting to the sheets
        requests_body = self._create_requests_body(
            required_metadata_fields,
            ordered_metadata_fields,
            json_schema,
            spreadsheet_id,
        )

        # Execute requests
        execute_google_api_requests(
            self.sheet_service,
            requests_body,
            service_type="batch_update",
            spreadsheet_id=spreadsheet_id,
        )

        # Setting up spreadsheet permissions (setup so that anyone with the link can edit)
        self._set_permissions(spreadsheet_id)

        # generating spreadsheet URL
        manifest_url = "https://docs.google.com/spreadsheets/d/" + spreadsheet_id
        return manifest_url

    def _gather_all_fields(self, fields, json_schema):
        """Gather all the attributes/fields to include as columns in the manifest.
        Args:
            fields(list[str]): fields/attributes to search
        Returns:
            required_metadata_fields(dict):
                keys: of all the fields/attributes that need to be added
                    to the manifest
                values(list[str]): valid values
        """
        # Get required fields
        required_metadata_fields = self._get_required_metadata_fields(
            json_schema, fields
        )
        # Add additioal dependencies
        required_metadata_fields = self._gather_dependency_requirements(
            json_schema, required_metadata_fields
        )

        # Add additional metadata as entries to columns
        required_metadata_fields = self._get_additional_metadata(
            required_metadata_fields
        )
        return required_metadata_fields

    def get_empty_manifest(self, json_schema_filepath=None):
        """Create an empty manifest using specifications from the
        json schema.
        Args:
            json_schema_filepath (str): path to json schema file
        Returns:
            manifest_url (str): url of the google sheet manifest.
        """
        spreadsheet_id = self._create_empty_manifest_spreadsheet(self.title)
        json_schema = self._get_json_schema(json_schema_filepath)

        required_metadata_fields = self._gather_all_fields(
            json_schema["properties"].keys(), json_schema
        )

        manifest_url = self._create_empty_gs(
            required_metadata_fields, json_schema, spreadsheet_id
        )
        return manifest_url

    def _get_missing_columns(self, wb_header, manifest_df):
        # get headers from sheet 
        manifest_df_header = manifest_df.columns

        # find missing columns present in existing manifest but missing in latest schema
        out_of_schema_columns = set(manifest_df_header) - set(wb_header)

        return out_of_schema_columns


    def set_dataframe_by_url(
        self, manifest_url: str, manifest_df: pd.DataFrame
    ) -> ps.Spreadsheet:
        """Update Google Sheets using given pandas DataFrame.
        Args:
            manifest_url (str): Google Sheets URL.
            manifest_df (pd.DataFrame): Data frame to "upload".
        Returns:
            ps.Spreadsheet: A Google Sheet object.
        """

        # authorize pygsheets to read from the given URL
        gc = ps.authorize(custom_credentials=self.creds)

        # open google sheets and extract first sheet
        sh = gc.open_by_url(manifest_url)
        wb = sh[0]

        # Handle scenario when existing manifest does not match new
        #       manifest template due to changes in the data model:
        #
        # the sheet column header reflect the latest schema
        # the existing manifest column-set may be outdated
        # ensure that, if missing, attributes from the latest schema are added to the
        # column-set of the existing manifest so that the user can modify their data if needed
        # to comply with the latest schema

        # get headers from existing manifest and sheet 
        wb_header = wb.get_row(1)
        manifest_df_header = manifest_df.columns

        # find missing columns in existing manifest
        new_columns = set(wb_header) - set(manifest_df_header)

        # clean empty columns if any are present (there should be none)
        # TODO: Remove this line once we start preventing empty column names
        if '' in new_columns:
            new_columns = new_columns.remove('')

        # find missing columns present in existing manifest but missing in latest schema
        out_of_schema_columns = set(manifest_df_header) - set(wb_header)

        # update existing manifest w/ missing columns, if any
        if new_columns:
            manifest_df = manifest_df.assign(
                **dict(zip(new_columns, len(new_columns) * [""]))
            )

        # sort columns in the updated manifest:
        # match latest schema order
        # move obsolete columns at the end
        manifest_df = manifest_df[self.sort_manifest_fields(manifest_df.columns)]
        manifest_df = manifest_df[[c for c in manifest_df if c not in out_of_schema_columns] + list(out_of_schema_columns)]
    
        # The following line sets `valueInputOption = "RAW"` in pygsheets
        sh.default_parse = False

        # update spreadsheet with given manifest starting at top-left cell
        wb.set_dataframe(manifest_df, (1, 1))

        # update validation rules (i.e. no validation rules) for out of schema columns, if any
        # TODO: similarly clear formatting for out of schema columns, if any
        num_out_of_schema_columns = len(out_of_schema_columns)
        if num_out_of_schema_columns > 0: 
            start_col = self._column_to_letter(len(manifest_df.columns) - num_out_of_schema_columns) # find start of out of schema columns
            end_col = self._column_to_letter(len(manifest_df.columns) + 1) # find end of out of schema columns
       
            wb.set_data_validation(start = start_col, end = end_col, condition_type = None)

        # set permissions so that anyone with the link can edit
        sh.share("", role="writer", type="anyone")

        return sh

    def get_dataframe_by_url(self, manifest_url: str) -> pd.DataFrame:
        """Retrieve pandas DataFrame from table in Google Sheets.

        Args:
            manifest_url (str): Google Sheets URL.

        Return:
            pd.DataFrame: Data frame corresponding to table in given URL.
        """

        # authorize pygsheets to read from the given URL
        gc = ps.authorize(custom_credentials=self.creds)

        # open google sheets and extract first sheet
        sh = gc.open_by_url(manifest_url)
        wb = sh[0]

        # get column headers and read it into a dataframe
        manifest_df = wb.get_as_df(hasHeader=True)

        # An empty column is sometimes included
        if "" in manifest_df:
            manifest_df.drop(columns=[""], inplace=True)

        return manifest_df

    def map_annotation_names_to_display_names(
        self, annotations: pd.DataFrame
    ) -> pd.DataFrame:
        """Update columns names to use display names for consistency.

        Args:
            annotations (pd.DataFrame): Annotations table.

        Returns:
            pd.DataFrame: Annotations table with updated column headers.
        """
        # Get list of attribute nodes from data model
        model_nodes = self.sg.se.get_nx_schema().nodes

        # Subset annotations to those appearing as a label in the model
        labels = filter(lambda x: x in model_nodes, annotations.columns)

        # Generate a dictionary mapping labels to display names
        label_map = {l: model_nodes[l]["displayName"] for l in labels}

        # Use the above dictionary to rename columns in question
        return annotations.rename(columns=label_map)

    def get_manifest_with_annotations(
        self, annotations: pd.DataFrame
    ) -> Tuple[ps.Spreadsheet, pd.DataFrame]:
        """Generate manifest, optionally with annotations (if requested).

        Args:
            annotations (pd.DataFrame): Annotations table (can be empty).

        Returns:
            Tuple[ps.Spreadsheet, pd.DataFrame]: Both the Google Sheet
            URL and the corresponding data frame is returned.
        """

        # Map annotation labels to display names to match manifest columns
        annotations = self.map_annotation_names_to_display_names(annotations)

        # Convert annotations table into dictionary, but maintain order
        annotations_dict_raw = annotations.to_dict(into=OrderedDict)
        annotations_dict = OrderedDict(
            (k, list(v.values())) for k, v in annotations_dict_raw.items()
        )

        # Needs to happen before get_empty_manifest() gets called
        self.additional_metadata = annotations_dict

        # Generate empty manifest using `additional_metadata`
        manifest_url = self.get_empty_manifest()
        manifest_df = self.get_dataframe_by_url(manifest_url)

        # Annotations clashing with manifest attributes are skipped
        # during empty manifest generation. For more info, search
        # for `additional_metadata` in `self.get_empty_manifest`.
        # Hence, the shared columns need to be updated separately.
        if self.is_file_based and self.use_annotations:
            # This approach assumes that `update_df` returns
            # a data frame whose columns are in the same order
            manifest_df = update_df(manifest_df, annotations)
            manifest_sh = self.set_dataframe_by_url(manifest_url, manifest_df)
            manifest_url = manifest_sh.url

        return manifest_url, manifest_df

    def export_sheet_to_excel(self, title: str = None, manifest_url : str = None, output_location: str = None) -> str:
        """
        export manifest as an Excel spreadsheet and return local file path
        Args:
            title: title of the exported excel spreadsheet
            manifest_url: manifest google sheet url 
            output_location: the location where the exported excel file would live
        return: 
            Export manifest to a desired location. 
        """
        # construct file name
        file_name = title + ".xlsx"
        #if file path exists and it contains a file name: 
        if output_location: 
            if os.path.exists(output_location):
                if Path(output_location).suffix == '.xlsx' or Path(output_location).suffix == '.xls':
                    output_excel_file_path = output_location
                # if users define the location but it doesn't contain a file name, we should add the file name:
                else:
                    output_excel_file_path = os.path.join(output_location, file_name)
    
        # trigger a warning if file path is provided but does not exist
        elif output_location and not os.path.exists(output_location):
            output_excel_file_path = os.path.abspath(os.path.join(os.getcwd(), file_name))
            logger.warning(f'{output_location} does not exist. Using current working directory {output_excel_file_path}')
        # otherwise, use the default location
        else:
            output_excel_file_path = os.path.abspath(os.path.join(os.getcwd(), file_name))
        # export the manifest to excel
        export_manifest_drive_service(manifest_url, file_path=output_excel_file_path, mimeType = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')
        
        return output_excel_file_path

    def _handle_output_format_logic(self, output_format: str = None, output_path: str = None, sheet_url: bool = None, manifest_url: str = None, empty_manifest_url: str = None, dataframe: pd.DataFrame = None):
        """
        handle the logic between sheet_url parameter and output_format parameter to determine the type of output to return
        Args: 
            output_format: Determines if Google sheet URL, pandas dataframe, or Excel spreadsheet gets returned.
            sheet_url (Will be deprecated): a boolean ; determine if a pandas dataframe or a google sheet url gets return
            manifest_url: Google sheet URL populated with metadata
            empty_manifest_url: Google sheet URL that leads to an empty manifest 
            dataframe: the pandas dataframe that contains the metadata that needs to be populated to an empty manifest
            output_path: Determines the output path of the exported manifest (only relevant if returning an excel spreadsheet)
        return: a pandas dataframe, file path of an excel spreadsheet, or a google sheet URL 
        """

        # check if output_format parameter gets set. If not, check the sheet_url parameter
        if not output_format: 
            # if the sheet_url parameter gets set to True, return a google sheet url 
            if sheet_url: 
                return manifest_url
                # else, return a pandas dataframe
            else: 
                return dataframe
        else:
            # if the output type gets set to "dataframe", return a data frame 
            if output_format == "dataframe":
                return dataframe
            # if the output type gets set to "excel", return an excel spreadsheet
            elif output_format == "excel":                 
                # export the manifest url to excel
                output_file_path = self.export_sheet_to_excel(title = self.title, manifest_url = empty_manifest_url, output_location = output_path)

                if not os.path.exists(output_file_path): 
                    logger.error(f'Export to Excel spreadsheet fail. Please make sure that file path {output_file_path} is valid')

                # populate an excel spreadsheet with the existing dataframe
                self.populate_existing_excel_spreadsheet(output_file_path, dataframe)

                return output_file_path
            # for all other cases, return a google sheet url 
            else: 
                return manifest_url  

    def get_manifest(
        self, dataset_id: str = None, sheet_url: bool = None, json_schema: str = None, output_format: str = None, output_path: str = None
    ) -> Union[str, pd.DataFrame]:
        """Gets manifest for a given dataset on Synapse.
           TODO: move this function to class MetadatModel (after MetadataModel is refactored)

        Args:
            dataset_id: Synapse ID of the "dataset" entity on Synapse (for a given center/project).
            sheet_url: Determines if googlesheet URL or pandas dataframe should be returned.
            output_format: Determines if Google sheet URL, pandas dataframe, or Excel spreadsheet gets returned.
            output_path: Determines the output path of the exported manifest

        Returns:
            Googlesheet URL, pandas dataframe, or an Excel spreadsheet 
        """

        # Handle case when no dataset ID is provided
        if not dataset_id:
            manifest_url = self.get_empty_manifest(json_schema_filepath=json_schema)

            # if output_form parameter is set to "excel", return an excel spreadsheet
            if output_format == "excel": 
                output_file_path = self.export_sheet_to_excel(title = self.title, manifest_url = manifest_url, output_location = output_path)
                return output_file_path
            # since we are not going to return an empty dataframe for an empty manifest, here we will just return a google sheet url for all other cases
            else: 
                return manifest_url

        # Otherwise, create manifest using the given dataset
        #TODO: avoid explicitly exposing Synapse store functionality
        # just instantiate a Store class and let it decide at runtime/config
        # the store type

        store = SynapseStorage()

        # Get manifest file associated with given dataset (if applicable)
        # populate manifest with set of new files (if applicable)
        manifest_record = store.updateDatasetManifestFiles(self.sg, datasetId = dataset_id, store = False)

        # get URL of an empty manifest file created based on schema component
        empty_manifest_url = self.get_empty_manifest()

        # Populate empty template with existing manifest
        if manifest_record:

            # TODO: Update or remove the warning in self.__init__() if
            # you change the behavior here based on self.use_annotations
                            
            # populate empty manifest with content from downloaded/existing manifest
            manifest_sh = self.set_dataframe_by_url(empty_manifest_url, manifest_record[1])

            # determine the format of manifest
            result = self._handle_output_format_logic(output_format = output_format, output_path = output_path, sheet_url = sheet_url, manifest_url = manifest_sh.url, empty_manifest_url=empty_manifest_url, dataframe = manifest_record[1])
            return result

        # Generate empty template and optionally fill in with annotations
        else:
            # Using getDatasetAnnotations() to retrieve file names and subset
            # entities to files and folders (ignoring tables/views)
            annotations = pd.DataFrame()
            if self.is_file_based:
                annotations = store.getDatasetAnnotations(dataset_id)

            # if there are no files with annotations just generate an empty manifest
            if annotations.empty:
                manifest_url = self.get_empty_manifest() 
                manifest_df = self.get_dataframe_by_url(manifest_url)
            else:
                # Subset columns if no interested in user-defined annotations and there are files present
                if self.is_file_based and not self.use_annotations:
                    annotations = annotations[["Filename", "eTag", "entityId"]]

                # Update `additional_metadata` and generate manifest
                manifest_url, manifest_df = self.get_manifest_with_annotations(annotations)
            
            # determine the format of manifest that gets return 
            result = self._handle_output_format_logic(output_format = output_format, output_path = output_path, sheet_url = sheet_url, manifest_url = manifest_url, empty_manifest_url=empty_manifest_url, dataframe = manifest_df)
            return result

    def populate_existing_excel_spreadsheet(self, existing_excel_path: str = None, additional_df: pd.DataFrame = None):
        '''Populate an existing excel spreadsheet by using an additional dataframe (to avoid sending metadata directly to Google APIs)
        Args:
            existing_excel_path: path of an existing excel spreadsheet
            additional_df: additional dataframe
        '''
        # load workbook
        workbook = load_workbook(existing_excel_path)

        #get missing columns and sort manifest 
        #get existing headers
        existing_df = pd.read_excel(existing_excel_path, sheet_name="Sheet1")
        workbook_headers = existing_df.columns

        # get new column headers
        out_of_schema_columns = self._get_missing_columns(workbook_headers, additional_df)
        out_of_schema_columns_lst = list(out_of_schema_columns)
        
        # initalize excel writer
        writer = pd.ExcelWriter(existing_excel_path, engine='openpyxl')
        writer.book = workbook
        writer.sheets = {ws.title: ws for ws in workbook.worksheets}
        worksheet = writer.sheets["Sheet1"]

        # add additional content to the existing spreadsheet
        additional_df.to_excel(writer, sheet_name = "Sheet1", startrow=1, index = False, header=False)

        # if there are new columns, add them to the end of spreadsheet
        if len(out_of_schema_columns_lst) > 0:
            for i, col_name in enumerate(out_of_schema_columns_lst):
                col_index = len(workbook_headers) + 1 + i
                worksheet.cell(row=1, column=col_index).value = col_name
        
        # save and close
        writer.save()
        writer.close()

    def populate_manifest_spreadsheet(self, existing_manifest_path: str = None, empty_manifest_url: str = None, return_excel: bool = False, title: str = None):
        """Creates a google sheet manifest based on existing manifest.
        Args:
            existing_manifest_path: the location of the manifest containing metadata presently stored
            empty_manifest_url: the path to a manifest template to be prepopulated with existing's manifest metadata
            return_excel: if true, return an Excel spreadsheet instead of Google sheet
            title: title of output manifest 
        """

        # read existing manifest
        manifest = load_df(existing_manifest_path)
        
        if return_excel: 
            '''if we are returning an Excel spreadsheet, do not populate dataframe to google'''
            # get an empty manifest 
            manifest_url = empty_manifest_url

            # export the manifest to excel
            output_excel_file_path = self.export_sheet_to_excel(manifest_url = manifest_url, title=title)
            # populate exported sheet 
            self.populate_existing_excel_spreadsheet(output_excel_file_path, manifest)
            return output_excel_file_path
        else: 
            manifest_sh = self.set_dataframe_by_url(empty_manifest_url, manifest)
            return manifest_sh.url

    def sort_manifest_fields(self, manifest_fields, order="schema"):
        # order manifest fields alphabetically (base order)
        manifest_fields = sorted(manifest_fields)

        if order == "alphabetical":
            # if the order is alphabetical ensure that filename is first, if present
            if "Filename" in manifest_fields:
                manifest_fields.remove("Filename")
                manifest_fields.insert(0, "Filename")

        # order manifest fields based on data-model schema
        if order == "schema":
            if self.sg and self.root:
                # get display names of dependencies
                dependencies_display_names = self.sg.get_node_dependencies(self.root)

                # reorder manifest fields so that root dependencies are first and follow schema order
                manifest_fields = sorted(
                    manifest_fields,
                    key=lambda x: dependencies_display_names.index(x)
                    if x in dependencies_display_names
                    else len(manifest_fields) - 1,
                )
            else:
                raise ValueError(
                    f"Provide valid data model path and valid component from data model."
                )

        # always have entityId as last columnn, if present
        if "entityId" in manifest_fields:
            manifest_fields.remove("entityId")
            manifest_fields.append("entityId")

        return manifest_fields



import logging

import pandas as pd
import numpy as np
from copy import deepcopy

logger = logging.getLogger(__name__)


def load_df(file_path, preserve_raw_input=True, data_model=False, **load_args):
    """
    Universal function to load CSVs and return DataFrames
    Args:
        file_path: path of csv to open
        preserve_raw_input: Bool. If false, convert cell datatypes to an inferred type
        data_model: bool, indicates if importing a data model
        load_args: dict of key value pairs to be passed to the pd.read_csv function
        **kwargs: keyword arguments for pd.read_csv()

    Returns: a processed dataframe for manifests or unprocessed df for data models
    """
    #Read CSV to df as type specified in kwargs
    org_df = pd.read_csv(file_path, keep_default_na = True, encoding='utf8', **load_args)
    
    if preserve_raw_input:
        #only trim if not data model csv
        if not data_model:
            org_df=trim_commas_df(org_df)

        return org_df

    else:
        float_df=deepcopy(org_df)
        #Find integers stored as strings
        #Cast the columns in dataframe to string while preserving NaN
        null_cells = org_df.isnull() 
        org_df = org_df.astype(str).mask(null_cells, '')
        ints = org_df.applymap(lambda x: np.int64(x) if str.isdigit(x) else False, na_action='ignore').fillna(False)

        #convert strings to numerical dtype (float) if possible, preserve non-numerical strings
        for col in org_df.columns:
            float_df[col]=pd.to_numeric(float_df[col], errors='coerce')
            float_df[col].fillna(org_df[col][float_df[col].isna()],inplace=True)
        
        #Trim nans and empty rows and columns
        processed_df = trim_commas_df(float_df)
        
        #Store values that were entered as ints
        processed_df=processed_df.mask(ints != False, other = ints)  

        return processed_df

def normalize_table(df: pd.DataFrame, primary_key: str) -> pd.DataFrame:

    """
    Function to normalize a table (e.g. dedup)
    Args:
        df: data frame to normalize
        primary_key: primary key on which to perform dedup

    Returns: a dedupped dataframe
    """

    try:
        # if valid primary key has been provided normalize df
        df = df.reset_index()
        df_norm = df.drop_duplicates(subset=[primary_key])
        df_norm = df.drop(columns=["index"])
        return df_norm
    except KeyError:
        # if the primary key is not in the df; then return the same df w/o changes
        logger.warning(
            "Specified primary key is not in table schema. Proceeding without table changes."
        )

        return df


def update_df(
    input_df: pd.DataFrame, updates_df: pd.DataFrame, index_col: str = "entityId"
) -> pd.DataFrame:
    """Update a manifest using another data frame with Synapse IDs.

    The input `input_df` is copied to avoid changing the input.

    Both input data frames must have an `entityId` column. Any rows
    in `updates_df` corresponding to entities that don't appear in
    `input_df` are silently dropped. Similarly, any columns in
    `updates_df` that don't appear in `input_df` are not added.

    IMPORTANT: This function is currently designed to handle empty
    manifests because it will not raise an error or warning if any
    overwriting of existing values takes place.

    TODO: Handle conflicts/overwriting more elegantly. See:
    https://github.com/Sage-Bionetworks/schematic/issues/312#issuecomment-725750931

    Args:
        input_df (pd.DataFrame): Manifest data frame. Must
        include the `index_col` column.
        updates_df (pd.DataFrame): Data frame with updates. Must
        include the `index_col` column. This data frame doesn't
        need to include all of the column names from `input_df`.
        index_col (str): Column to index the data frames on.

    Returns:
        pd.DataFrame: Updated `input_df` data frame.
    """
    # Confirm that entityId is present in both data frames
    assert index_col in input_df, f"`input_df` lacks `{index_col}` column."
    assert index_col in updates_df, f"`updates_df` lacks `{index_col}` column."

    # Set `inplace=False` to copy input data frames and avoid side-effects
    input_df_idx = input_df.set_index(index_col, inplace=False)
    updates_df_idx = updates_df.set_index(index_col, inplace=False)

    # Update manifest data frame and reset index
    input_df_idx.update(updates_df_idx, overwrite=True)

    # Undo index and ensure original column order
    input_df_idx.reset_index(inplace=True)
    input_df_idx = input_df_idx[input_df.columns]

     # Sometimes pandas update can change the column datatype, recast
    for col in input_df_idx.columns:
        input_df_idx[col] = input_df_idx[col].astype(input_df.dtypes[col])

    return input_df_idx


def trim_commas_df(df: pd.DataFrame):
    """Removes empty (trailing) columns and empty rows from pandas dataframe (manifest data).

    Args:
        df: pandas dataframe with data from manifest file.

    Returns:
        df: cleaned-up pandas dataframe.
    """
    # remove all columns which have substring "Unnamed" in them
    df = df.loc[:, ~df.columns.str.contains("^Unnamed")]

    # remove all completely empty rows
    df = df.dropna(how="all", axis=0)

    #Fill in nan cells with empty strings
    df.fillna("", inplace=True)
    return df



import json
from statistics import mode
from tabnanny import check
from jsonschema import Draft7Validator, exceptions, ValidationError
import logging

import numpy as np
import os
import pandas as pd
import re
import sys

# allows specifying explicit variable types
from typing import Any, Dict, Optional, Text, List
from urllib.parse import urlparse
from urllib.request import urlopen, OpenerDirector, HTTPDefaultErrorHandler
from urllib.request import Request
from urllib import error

from schematic.models.validate_attribute import ValidateAttribute, GenerateError
from schematic.schemas.generator import SchemaGenerator
from schematic.store.synapse import SynapseStorage
from schematic.models.GE_Helpers import GreatExpectationsHelpers
from schematic.utils.validate_rules_utils import validation_rule_info
from schematic.utils.validate_utils import rule_in_rule_list

logger = logging.getLogger(__name__)

class ValidateManifest(object):
    def __init__(self, errors, manifest, manifestPath, sg, jsonSchema):
        self.errors = errors
        self.manifest = manifest
        self.manifestPath = manifestPath
        self.sg = sg
        self.jsonSchema = jsonSchema       

    def get_multiple_types_error(
        self, validation_rules: list, attribute_name: str, error_type: str
    ) -> List[str]:
        """
            Generate error message for errors when trying to specify 
            multiple validation rules.
            """
        error_col = attribute_name  # Attribute name
        if error_type == "too_many_rules":
            error_str = (
                f"For attribute {attribute_name}, the provided validation rules ({validation_rules}) ."
                f"have too many entries. We currently only specify two rules ('list :: another_rule')."
            )
            logging.error(error_str)
            error_message = error_str
            error_val = f"Multiple Rules: too many rules"
        if error_type == "list_not_first":
            error_str = (
                f"For attribute {attribute_name}, the provided validation rules ({validation_rules}) are improperly "
                f"specified. 'list' must be first."
            )
            logging.error(error_str)
            error_message = error_str
            error_val = f"Multiple Rules: list not first"
        return ["NA", error_col, error_message, error_val]

    def validate_manifest_rules(
        self, manifest: pd.core.frame.DataFrame, sg: SchemaGenerator, restrict_rules: bool, project_scope: List,
    ) -> (pd.core.frame.DataFrame, List[List[str]]):
        """
        Purpose:
            Take validation rules set for a particular attribute
            and validate manifest entries based on these rules.
        Input:
            manifest: pd.core.frame.DataFrame
                imported from models/metadata.py
                contains metadata input from user for each attribute.
            sg: SchemaGenerator
                initialized within models/metadata.py
        Returns:
            manifest: pd.core.frame.DataFrame
                If a 'list' validatior is run, the manifest needs to be 
                updated to change the attribute column values to a list.
                In this case the manifest will be updated then exported.
            errors: List[List[str]]
                If any errors are generated they will be added to an errors
                list log recording the following information:
                [error_row, error_col, error_message, error_val]
        TODO: 
            -Investigate why a :: delimiter is breaking up the
                validation rules without me having to do anything...
            - Move the rules formatting validation to the JSONLD 
                generation script.
        """

        # for each type of rule that can be spefified (key) point
        # to the type of validation that will be run.

        validation_types = validation_rule_info()

        type_dict={
            "float64": float,
            "int64": int,
            "str": str,
        }

        unimplemented_expectations=[
            "url",
            "list",
            "regex.*",
            "matchAtLeastOne.*",
            "matchExactlyOne.*",
            ]

        in_house_rules = [
            "int",
            "float",
            "num",
            "str",
            "regex.*",
            "url",
            "list",
            "matchAtLeastOne.*",
            "matchExactlyOne.*",
        ]

        # initialize error and warning handling lists.
        errors = []   
        warnings = [] 

        if not restrict_rules:
            #operations necessary to set up and run ge suite validation
            ge_helpers=GreatExpectationsHelpers(
                sg=sg,
                unimplemented_expectations=unimplemented_expectations,
                manifest = manifest,
                manifestPath = self.manifestPath,
                )

            ge_helpers.build_context()
            ge_helpers.build_expectation_suite()
            ge_helpers.build_checkpoint()

        #run GE validation
            results = ge_helpers.context.run_checkpoint(
                checkpoint_name="manifest_checkpoint",
                batch_request={
                    "runtime_parameters": {"batch_data": manifest},
                    "batch_identifiers": {
                        "default_identifier_name": "manifestID"
                    },
                },
                result_format={'result_format': 'COMPLETE'},
            )        
        
            #print(results)       
            #results.list_validation_results()
            validation_results = results.list_validation_results()
            

            #parse validation results dict and generate errors
            errors, warnings = ge_helpers.generate_errors(
                errors = errors,
                warnings = warnings,
                validation_results = validation_results,
                validation_types = validation_types,
                sg = sg,
                )               
        else:             
            logging.info("Great Expetations suite will not be utilized.")  


        regex_re=re.compile('regex.*')
        for col in manifest.columns:
            # remove trailing/leading whitespaces from manifest
            manifest.applymap(lambda x: x.strip() if isinstance(x, str) else x)
            validation_rules = sg.get_node_validation_rules(col)

            # Check that attribute rules conform to limits:
            # no more than two rules for an attribute. 
            # As more combinations get added, may want to bring out into its own function / or use validate_rules_utils?
            if len(validation_rules) > 2:
                errors.append(
                    self.get_multiple_types_error(
                        validation_rules, col, error_type="too_many_rules"
                    )
                )

            # Given a validation rule, run validation. Skip validations already performed by GE
            for rule in validation_rules:
                validation_type = rule.split(" ")[0]
                if rule_in_rule_list(rule,unimplemented_expectations) or (rule_in_rule_list(rule,in_house_rules) and restrict_rules):
                    if not rule_in_rule_list(rule,in_house_rules):
                        logging.warning(f"Validation rule {rule.split(' ')[0]} has not been implemented in house and cannnot be validated without Great Expectations.")
                        continue  

                    #Validate for each individual validation rule.
                    validation_method = getattr(
                            ValidateAttribute, validation_types[validation_type]['type']
                        )

                    if validation_type == "list":
                        vr_errors, vr_warnings, manifest_col = validation_method(
                            self, rule, manifest[col], sg,
                        )
                        manifest[col] = manifest_col
                    elif validation_type.lower().startswith("match"):
                        vr_errors, vr_warnings = validation_method(
                            self, rule, manifest[col], project_scope, sg,
                        )
                    else:
                        vr_errors, vr_warnings = validation_method(
                            self, rule, manifest[col], sg,
                        )
                    # Check for validation rule errors and add them to other errors.
                    if vr_errors:
                        errors.extend(vr_errors)
                    if vr_warnings:
                        warnings.extend(vr_warnings)

        return manifest, errors, warnings

    def validate_manifest_values(self, manifest, jsonSchema, sg
    ) -> (List[List[str]], List[List[str]]):
        
        errors = []
        warnings = []
        col_attr = {} # save the mapping between column index and attribute name
        
        # numerical values need to be type string for the jsonValidator
        for col in manifest.select_dtypes(include=[int, np.int64, float, np.float64]).columns:
            manifest[col]=manifest[col].astype('string')
        manifest = manifest.applymap(lambda x: str(x) if isinstance(x, (int, np.int64, float, np.float64)) else x, na_action='ignore')

        annotations = json.loads(manifest.to_json(orient="records"))
        for i, annotation in enumerate(annotations):
            v = Draft7Validator(jsonSchema)
            for error in sorted(v.iter_errors(annotation), key=exceptions.relevance):
                errorRow = i + 2
                errorCol = error.path[-1] if len(error.path) > 0 else "Wrong schema"
                errorColName = error.path[0] if len(error.path) > 0 else "Wrong schema"
                errorMsg = error.message[0:500]
                errorVal = error.instance if len(error.path) > 0 else "Wrong schema"

                val_errors, val_warnings =  GenerateError.generate_schema_error(row_num = errorRow, attribute_name = errorColName, error_msg = errorMsg, invalid_entry = errorVal, sg = sg)

                if val_errors:
                    errors.append(val_errors)
                if val_warnings:
                    warnings.append(val_warnings)

        return errors, warnings


def validate_all(self, errors, warnings, manifest, manifestPath, sg, jsonSchema, restrict_rules, project_scope: List):
    vm = ValidateManifest(errors, manifest, manifestPath, sg, jsonSchema)
    manifest, vmr_errors, vmr_warnings = vm.validate_manifest_rules(manifest, sg, restrict_rules, project_scope)
    if vmr_errors:
        errors.extend(vmr_errors)
    if vmr_warnings:
        warnings.extend(vmr_warnings)

    vmv_errors, vmv_warnings = vm.validate_manifest_values(manifest, jsonSchema, sg)
    if vmv_errors:
        errors.extend(vmv_errors)
    if vmv_warnings:
        warnings.extend(vmv_warnings)

    return errors, warnings, manifest



import os
from jsonschema import validate

from schematic.utils.io_utils import load_schemaorg, load_json, load_default
from schematic.utils.general import str2list, dict2list, find_duplicates
from schematic.utils.curie_utils import (
    expand_curies_in_schema,
    extract_name_from_uri_or_curie,
)
from schematic.utils.validate_utils import (
    validate_class_schema,
    validate_property_schema,
    validate_schema,
)

from schematic import CONFIG


class SchemaValidator:
    """Validate Schema against SchemaOrg standard

    Validation Criterias:
    1. Data Structure wise:
      > "@id", "@context", "@graph"
      > Each element in "@graph" should contain "@id", "@type", "rdfs:comment",
      "rdfs:label", "sms:displayName"
      > validate against JSON Schema
      > Should validate the whole structure, and also validate property and
      value separately
    2. Data Content wise:
      > "@id" field should match with "rdfs:label" field
      > all prefixes used in the file should be defined in "@context"
      > There should be no duplicate "@id"
      > Class specific
        > rdfs:label field should be capitalize the first character of each
          word for a class;
        > the value of "rdfs:subClassOf" should be present in the schema or in
          the core vocabulary
        > sms:displayName ideally should contain capitalized words separated by space, but that's not enforced by validation
      > Property specific
        > rdfs:label field should be cammelCase
        > the value of "schema:domainIncludes" should be present in the schema
          or in the core vocabulary
        > the value of "schema:rangeIncludes" should be present in the schema
          or in the core vocabulary
        > sms:displayName ideally should contain capitalized words separated by space, but that's not enforced by validation
        TODO: add dependencies and component dependencies to class structure documentation; as well as value range and required property

    """

    def __init__(self, schema):
        self.schemaorg = {"schema": load_schemaorg(), "classes": [], "properties": []}
        for _schema in self.schemaorg["schema"]["@graph"]:
            for _record in _schema["@graph"]:
                if "@type" in _record:
                    _type = str2list(_record["@type"])
                    if "rdfs:Property" in _type:
                        self.schemaorg["properties"].append(_record["@id"])
                    elif "rdfs:Class" in _type:
                        self.schemaorg["classes"].append(_record["@id"])
        self.extension_schema = {
            "schema": expand_curies_in_schema(schema),
            "classes": [],
            "properties": [],
        }
        for _record in self.extension_schema["schema"]["@graph"]:
            _type = str2list(_record["@type"])
            if "rdfs:Property" in _type:
                self.extension_schema["properties"].append(_record["@id"])
            elif "rdfs:Class" in _type:
                self.extension_schema["classes"].append(_record["@id"])
        self.all_classes = self.schemaorg["classes"] + self.extension_schema["classes"]

    def validate_class_label(self, label_uri):
        """Check if the first character of class label is capitalized"""
        label = extract_name_from_uri_or_curie(label_uri)
        assert label[0].isupper()

    def validate_property_label(self, label_uri):
        """Check if the first character of property label is lower case"""
        label = extract_name_from_uri_or_curie(label_uri)
        assert label[0].islower()

    def validate_subclassof_field(self, subclassof_value):
        """Check if the value of "subclassof" is included in the schema file"""
        subclassof_value = dict2list(subclassof_value)
        for record in subclassof_value:
            assert record["@id"] in self.all_classes

    def validate_domainIncludes_field(self, domainincludes_value):
        """Check if the value of "domainincludes" is included in the schema
        file
        """
        domainincludes_value = dict2list(domainincludes_value)
        for record in domainincludes_value:
            assert record["@id"] in self.all_classes, (
                "value of domainincludes not recorded in schema: %r"
                % domainincludes_value
            )

    def validate_rangeIncludes_field(self, rangeincludes_value):
        """Check if the value of "rangeincludes" is included in the schema
        file
        """
        rangeincludes_value = dict2list(rangeincludes_value)
        for record in rangeincludes_value:
            assert record["@id"] in self.all_classes

    def check_whether_atid_and_label_match(self, record):
        """Check if @id field matches with the "rdfs:label" field"""
        _id = extract_name_from_uri_or_curie(record["@id"])
        assert _id == record["rdfs:label"], "id and label not match: %r" % record

    def check_duplicate_labels(self):
        """Check for duplication in the schema"""
        labels = [
            _record["rdfs:label"]
            for _record in self.extension_schema["schema"]["@graph"]
        ]
        duplicates = find_duplicates(labels)
        try:
            assert len(duplicates) == 0
        except:
            raise Exception("Duplicates detected in graph: ", duplicates)

    def validate_schema(self, schema):
        """Validate schema against SchemaORG standard"""
        json_schema_path = os.path.join("validation_schemas", "schema.json")
        json_schema = load_json(json_schema_path)
        return validate(schema, json_schema)

    def validate_property_schema(self, schema):
        """Validate schema against SchemaORG property definition standard"""
        json_schema_path = os.path.join(
            "validation_schemas", "property_json_schema.json"
        )
        json_schema = load_json(json_schema_path)
        return validate(schema, json_schema)

    def validate_class_schema(self, schema):
        """Validate schema against SchemaORG class definition standard"""
        json_schema_path = os.path.join("validation_schemas", "class_json_schema.json")
        json_schema = load_json(json_schema_path)
        return validate(schema, json_schema)

    def validate_full_schema(self):
        self.check_duplicate_labels()
        for record in self.extension_schema["schema"]["@graph"]:
            self.check_whether_atid_and_label_match(record)
            if record["@type"] == "rdf:Class":
                self.validate_class_schema(record)
                self.validate_class_label(record["@id"])
            elif record["@type"] == "rdf:Property":
                self.validate_property_schema(record)
                self.validate_property_label(record["@id"])
                self.validate_domainIncludes_field(
                    record["http://schema.org/domainIncludes"]
                )
                if "http://schema.org/rangeIncludes" in record:
                    self.validate_rangeIncludes_field(
                        record["http://schema.org/rangeIncludes"]
                    )



#!/usr/bin/env python3

from gc import callbacks
import logging
import sys

import click
import click_log

from jsonschema import ValidationError

from schematic.models.metadata import MetadataModel
from schematic.utils.cli_utils import get_from_config, fill_in_from_config, query_dict, parse_synIDs, parse_comma_str_to_list
from schematic.help import model_commands
from schematic.exceptions import MissingConfigValueError
from schematic import CONFIG

logger = logging.getLogger(__name__)
click_log.basic_config(logger)

CONTEXT_SETTINGS = dict(help_option_names=["--help", "-h"])  # help options

# invoke_without_command=True -> forces the application not to show aids before losing them with a --h
@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-c",
    "--config",
    type=click.Path(),
    envvar="SCHEMATIC_CONFIG",
    help=query_dict(model_commands, ("model", "config")),
)
@click.pass_context
def model(ctx, config):  # use as `schematic model ...`
    """
    Sub-commands for Metadata Model related utilities/methods.
    """
    try:
        logger.debug(f"Loading config file contents in '{config}'")
        ctx.obj = CONFIG.load_config(config)
    except ValueError as e:
        logger.error("'--config' not provided or environment variable not set.")
        logger.exception(e)
        sys.exit(1)


# prototype based on submit_metadata_manifest()
@model.command(
    "submit", short_help=query_dict(model_commands, ("model", "submit", "short_help"))
)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-mp",
    "--manifest_path",
    help=query_dict(model_commands, ("model", "submit", "manifest_path")),
)
@click.option(
    "-d",
    "--dataset_id",
    help=query_dict(model_commands, ("model", "submit", "dataset_id")),
)
@click.option(
    "-vc",
    "--validate_component",
    help=query_dict(model_commands, ("model", "submit", "validate_component")),
)
@click.option(
    "--use_schema_label/--use_display_label",
    "-sl/-dl",
    default=True,
    help=query_dict(model_commands, ("model", "submit", "use_schema_label")),
)
@click.option(
    "--hide_blanks",
    "-hb",
    is_flag=True,
    help=query_dict(model_commands,("model","submit","hide_blanks")),
)
@click.option(
    "--manifest_record_type",
    "-mrt",
    default='both',
    type=click.Choice(['table', 'entity', 'both'], case_sensitive=True),
    help=query_dict(model_commands, ("model", "submit", "manifest_record_type")))
@click.option(
    "-rr",
    "--restrict_rules",
    is_flag=True,
    help=query_dict(model_commands,("model","validate","restrict_rules")),
)
@click.option(
    "-ps",
    "--project_scope",
    default=None,
    callback=parse_synIDs,
    help=query_dict(model_commands, ("model", "validate", "project_scope")),
)
@click.pass_obj
def submit_manifest(
    ctx, manifest_path, dataset_id, validate_component, manifest_record_type, use_schema_label, hide_blanks, restrict_rules, project_scope,
):
    """
    Running CLI with manifest validation (optional) and submission options.
    """

    jsonld = get_from_config(CONFIG.DATA, ("model", "input", "location"))

    model_file_type = get_from_config(CONFIG.DATA, ("model", "input", "file_type"))

    metadata_model = MetadataModel(
        inputMModelLocation=jsonld, inputMModelLocationType=model_file_type
    )

    try:
        manifest_id = metadata_model.submit_metadata_manifest(
            path_to_json_ld = jsonld,
            manifest_path=manifest_path,
            dataset_id=dataset_id,
            validate_component=validate_component,
            manifest_record_type=manifest_record_type,
            restrict_rules=restrict_rules,
            use_schema_label=use_schema_label,
            hide_blanks=hide_blanks,
            project_scope=project_scope,
        )

        '''
        if censored_manifest_id:
            logger.info(
                f"File at '{manifest_path}' was censored and successfully associated "
                f"with dataset '{dataset_id}'. "
                f"An uncensored version has also been associated with dataset '{dataset_id}' "
                f"and submitted to the Synapse Access Control Team to begin the process "
                f"of adding terms of use or review board approval."
            )
        '''
        if manifest_id:
            logger.info(
                f"File at '{manifest_path}' was successfully associated "
                f"with dataset '{dataset_id}'."
            )
    except ValueError:
        logger.error(
            f"Component '{validate_component}' is not present in '{jsonld}', or is invalid."
        )
    except ValidationError:
        logger.error(
            f"Validation errors resulted while validating with '{validate_component}'."
        )
    except LookupError:
        logger.error(
            f"'{dataset_id}' could not be found in the asset view (or file view for Synapse user)"
        )


# prototype based on validateModelManifest()
@model.command(
    "validate",
    short_help=query_dict(model_commands, ("model", "validate", "short_help")),
)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-mp",
    "--manifest_path",
    type=click.Path(exists=True),
    required=True,
    help=query_dict(model_commands, ("model", "validate", "manifest_path")),
)
@click.option(
    "-dt",
    "--data_type",
    callback=parse_comma_str_to_list,
    help=query_dict(model_commands, ("model", "validate", "data_type")),
)
@click.option(
    "-js",
    "--json_schema",
    help=query_dict(model_commands, ("model", "validate", "json_schema")),
)
@click.option(
    "-rr",
    "--restrict_rules",
    is_flag=True,
    help=query_dict(model_commands,("model","validate","restrict_rules")),
)
@click.option(
    "-ps",
    "--project_scope",
    default=None,
    callback=parse_synIDs,
    help=query_dict(model_commands, ("model", "validate", "project_scope")),
)
@click.pass_obj
def validate_manifest(ctx, manifest_path, data_type, json_schema, restrict_rules,project_scope):
    """
    Running CLI for manifest validation.
    """
    data_type = fill_in_from_config("data_type", data_type, ("manifest", "data_type"))
    
    try:
        len(data_type) == 1
    except:
        logger.error(
            f"Can only validate a single data_type at a time. Please provide a single data_type"
        )

    data_type = data_type[0]

    json_schema = fill_in_from_config(
        "json_schema",
        json_schema,
        ("model", "input", "validation_schema"),
        allow_none=True,
    )
    
    jsonld = get_from_config(CONFIG.DATA, ("model", "input", "location"))

    model_file_type = get_from_config(CONFIG.DATA, ("model", "input", "file_type"))

    metadata_model = MetadataModel(
        inputMModelLocation=jsonld, inputMModelLocationType=model_file_type
    )

    errors, warnings = metadata_model.validateModelManifest(
        manifestPath=manifest_path, rootNode=data_type, jsonSchema=json_schema, restrict_rules=restrict_rules, project_scope=project_scope,
    )

    if not errors:
        click.echo(
            "Your manifest has been validated successfully. "
            "There are no errors in your manifest, and it can "
            "be submitted without any modifications."
        )
    else:
        click.echo(errors)



import builtins
import logging
import re
import sys
import time
from os import getenv
# allows specifying explicit variable types
from typing import Any, Dict, List, Optional, Text
from urllib import error
from urllib.parse import urlparse
from urllib.request import (HTTPDefaultErrorHandler, OpenerDirector, Request,
                            urlopen)

import numpy as np
import pandas as pd
from jsonschema import ValidationError

from schematic.schemas.generator import SchemaGenerator
from schematic.store.base import BaseStorage
from schematic.store.synapse import SynapseStorage
from schematic.utils.validate_rules_utils import validation_rule_info
from schematic.utils.validate_utils import (comma_separated_list_regex,
                                            parse_str_series_to_list)

logger = logging.getLogger(__name__)

class GenerateError:
    def generate_schema_error(row_num: str, attribute_name: str, error_msg: str, invalid_entry: str, sg: SchemaGenerator,)-> List[str]:
        '''
        Purpose: Process error messages generated from schema
        Input:
            - row_num: the row the error occurred on.
            - attribute_name: the attribute the error occurred on.
            - error_msg: Error message
        '''
        error_list = []
        warning_list = []
        
        #Determine which, if any, message to raise
        raises = GenerateError.get_message_level(
            val_rule = 'schema',
            attribute_name = attribute_name,
            sg = sg,
            )

        #if a message needs to be raised, get the approrpiate function to do so
        if raises:
            logLevel = getattr(logging,raises)  
        else:
            return error_list, warning_list


        error_col = attribute_name  # Attribute name
        error_row = row_num  # index row of the manifest where the error presented.
        error_message = error_msg

        arg_error_string = (
                f"For the attribute '{error_col}', on row {error_row}, {error_message}."
        )
        logLevel(arg_error_string)

        if raises == 'error':
            error_list = [error_row, error_col, error_message, invalid_entry]
        elif raises == 'warning':
            warning_list = [error_row, error_col, error_message, invalid_entry]

        return error_list, warning_list 

    def generate_list_error(
        list_string: str, row_num: str, attribute_name: str, list_error: str,
        invalid_entry:str, sg: SchemaGenerator, val_rule: str,
    ) -> List[str]:
        """
            Purpose:
                If an error is found in the string formatting, detect and record
                an error message.
            Input:
                - list_string: the user input list, that is represented as a string.
                - row_num: the row the error occurred on.
                - attribute_name: the attribute the error occurred on.
            Returns:
                Logging.error.
                Errors: List[str] Error details for further storage.
            """

        error_list = []
        warning_list = []
        
        #Determine which, if any, message to raise
        raises = GenerateError.get_message_level(
            val_rule = val_rule,
            attribute_name = attribute_name,
            sg = sg,
            )

        #if a message needs to be raised, get the approrpiate function to do so
        if raises:
            logLevel = getattr(logging,raises)  
        else:
            return error_list, warning_list

        if list_error == "not_comma_delimited":
            error_str = (
                f"For attribute {attribute_name} in row {row_num} it does not "
                f"appear as if you provided a comma delimited string. Please check "
                f"your entry ('{list_string}'') and try again."
            )
            logLevel(error_str)
            error_row = row_num  # index row of the manifest where the error presented.
            error_col = attribute_name  # Attribute name
            error_message = error_str
            error_val = invalid_entry
                #return error and empty list for warnings
        
        if raises == 'error':
            error_list = [error_row, error_col, error_message, error_val]
        #return warning and empty list for errors
        elif raises == 'warning':
            warning_list = [error_row, error_col, error_message, error_val]
        
        return error_list, warning_list 

    def generate_regex_error(
        val_rule: str,
        reg_expression: str,
        row_num: str,
        module_to_call: str,
        attribute_name: str,
        invalid_entry: str,
        sg: SchemaGenerator,
    ) -> List[str]:
        """
            Purpose:
                Generate an logging error as well as a stored error message, when
                a regex error is encountered.
            Input:
                val_rule: str, defined in the schema.
                reg_expression: str, defined in the schema
                row_num: str, row where the error was detected
                module_to_call: re module specified in the schema
                attribute_name: str, attribute being validated
            Returns:
                Logging.error.
                Errors: List[str] Error details for further storage.
            """
        error_list = []
        warning_list = []
        
        #Determine which, if any, message to raise
        raises = GenerateError.get_message_level(
            val_rule = val_rule,
            attribute_name = attribute_name,
            sg = sg,
            )

        #if a message needs to be raised, get the approrpiate function to do so
        if raises:
            logLevel = getattr(logging,raises)  
        else:
            return error_list, warning_list

        regex_error_string = (
            f"For the attribute {attribute_name}, on row {row_num}, the string is not properly formatted. "
            f'It should follow the following re.{module_to_call} pattern "{reg_expression}".'
        )
        logLevel(regex_error_string)
        error_row = row_num  # index row of the manifest where the error presented.
        error_col = attribute_name  # Attribute name
        error_message = regex_error_string
        error_val = invalid_entry

        #return error and empty list for warnings
        if raises == 'error':
            error_list = [error_row, error_col, error_message, error_val]
        #return warning and empty list for errors
        elif raises == 'warning':
            warning_list = [error_row, error_col, error_message, error_val]
        
        return error_list, warning_list       

    def generate_type_error(
        val_rule: str, row_num: str, attribute_name: str, invalid_entry:str, sg: SchemaGenerator,
    ) -> List[str]:
        """
            Purpose:
                Generate an logging error as well as a stored error message, when
                a type error is encountered.
            Input:
                val_rule: str, defined in the schema.
                row_num: str, row where the error was detected
                attribute_name: str, attribute being validated
            Returns:
                Logging.error.
                Errors: List[str] Error details for further storage.
            """

        error_list = []
        warning_list = []
        
        #Determine which, if any, message to raise
        raises = GenerateError.get_message_level(
            val_rule = val_rule,
            attribute_name = attribute_name,
            sg = sg,
            )

        #if a message needs to be raised, get the approrpiate function to do so
        if raises:
            logLevel = getattr(logging,raises)  
        else:
            return error_list, warning_list

        type_error_str = (
            f"On row {row_num} the attribute {attribute_name} "
            f"does not contain the proper value type {val_rule}."
        )
        logLevel(type_error_str)
        error_row = row_num  # index row of the manifest where the error presented.
        error_col = attribute_name  # Attribute name
        error_message = type_error_str
        error_val = invalid_entry


        #return error and empty list for warnings
        if raises == 'error':
            error_list = [error_row, error_col, error_message, error_val]
        #return warning and empty list for errors
        elif raises == 'warning':
            warning_list = [error_row, error_col, error_message, error_val]
        
        return error_list, warning_list              

    def generate_url_error(
        url: str, url_error: str, row_num: str, attribute_name: str, argument: str,
        invalid_entry:str, sg: SchemaGenerator, val_rule: str,
    ) -> List[str]:
        """
            Purpose:
                Generate an logging error as well as a stored error message, when
                a URL error is encountered.

                Types of errors included:
                    - Invalid URL: Refers to a URL that brings up an error when 
                        attempted to be accessed such as a HTTPError 404 Webpage Not Found.
                    - Argument Error: this refers to a valid URL that does not 
                        contain within it the arguments specified by the schema,
                        such as 'protocols.io' or 'dox.doi.org'
                    - Random Entry: this refers to an entry try that is not 
                        validated to be a URL.
                        e.g. 'lkejrlei', '0', 'not applicable'
            Input:
                url: str, that was input by the user.
                url_error: str, error detected in url_validation()
                attribute_name: str, attribute being validated
                argument: str, argument being validated.
            Returns:
                Logging.error.
                Errors: List[str] Error details for further storage.
            """

        error_list = []
        warning_list = []
        
        #Determine which, if any, message to raise
        raises = GenerateError.get_message_level(
            val_rule = val_rule,
            attribute_name = attribute_name,
            sg = sg,
            )

        #if a message needs to be raised, get the approrpiate function to do so
        if raises:
            logLevel = getattr(logging,raises)  
        else:
            return error_list, warning_list


        error_row = row_num  # index row of the manifest where the error presented.
        error_col = attribute_name  # Attribute name
        if url_error == "invalid_url":
            invalid_url_error_string = (
                f"For the attribute '{attribute_name}', on row {row_num}, the URL provided ({url}) does not "
                f"conform to the standards of a URL. Please make sure you are entering a real, working URL "
                f"as required by the Schema."
            )
            logLevel(invalid_url_error_string)
            error_message = invalid_url_error_string
            error_val = invalid_entry
        elif url_error == "arg_error":
            arg_error_string = (
                f"For the attribute '{attribute_name}', on row {row_num}, the URL provided ({url}) does not "
                f"conform to the schema specifications and does not contain the required element: {argument}."
            )
            logLevel(arg_error_string)
            error_message = arg_error_string
            error_val = f"URL Error: Argument Error"
        elif url_error == "random_entry":
            random_entry_error_str = (
                f"For the attribute '{attribute_name}', on row {row_num}, the input provided ('{url}'') does not "
                f"look like a URL, please check input and try again."
            )
            logLevel(random_entry_error_str)
            error_message = random_entry_error_str
            error_val = f"URL Error: Random Entry"

        #return error and empty list for warnings
        if raises == 'error':
            error_list = [error_row, error_col, error_message, error_val]
        #return warning and empty list for errors
        elif raises == 'warning':
            warning_list = [error_row, error_col, error_message, error_val]
        
        return error_list, warning_list        

    def generate_cross_warning(
        val_rule: str,
        attribute_name: str,
        sg: SchemaGenerator,
        matching_manifests = [],
        missing_manifest_ID = None,
        invalid_entry = None,
        row_num = None,
        
    ) -> List[str]:
        """
            Purpose:
                Generate an logging error as well as a stored error message, when
                a cross validation error is encountered.
            Input:
                val_rule: str, defined in the schema.
                matching_manifests: list of manifests with all values in the target attribute present
                manifest_ID: str, synID of the target manifest missing the source value
                attribute_name: str, attribute being validated
                invalid_entry: str, value present in source manifest that is missing in the target
                row_num: row in source manifest with value missing in target manifests             
            Returns:
                Logging.error.
                Errors: List[str] Error details for further storage.
            """
        error_list = []
        warning_list = []
        
        #Determine which, if any, message to raise
        raises = GenerateError.get_message_level(
            val_rule = val_rule,
            attribute_name = attribute_name,
            sg = sg,
            )

        #if a message needs to be raised, get the approrpiate function to do so
        if raises:
            logLevel = getattr(logging,raises)  
        else:
            return error_list, warning_list

        if val_rule.__contains__('matchAtLeast'):
            cross_error_str = (
                f"Value(s) {invalid_entry} from row(s) {row_num} of the attribute {attribute_name} in the source manifest are missing." )
            cross_error_str += f" Manifest(s) {missing_manifest_ID} are missing the value(s)." if missing_manifest_ID else ""
            
        elif val_rule.__contains__('matchExactly'):
            if matching_manifests != []:
                cross_error_str = (
                    f"All values from attribute {attribute_name} in the source manifest are present in {len(matching_manifests)} manifests instead of only 1.")
                cross_error_str += f" Manifests {matching_manifests} match the values in the source attribute." if matching_manifests else ""
                    
            elif val_rule.__contains__('set'):
                cross_error_str = (
                    f"No matches for the values from attribute {attribute_name} in the source manifest are present in any other manifests instead of being present in exactly 1. "
                )
            elif val_rule.__contains__('value'):
                cross_error_str = (
                    f"Value(s) {invalid_entry} from row(s) {row_num} of the attribute {attribute_name} in the source manifest are not present in only one other manifest. " 
                )            

        logLevel(cross_error_str)
        error_row = row_num  # index row of the manifest where the error presented.
        error_col = attribute_name  # Attribute name
        error_message = cross_error_str
        error_val = invalid_entry #Value from source manifest missing from targets
        
        #return error and empty list for warnings
        if raises == 'error':
            error_list = [error_row, error_col, error_message, error_val]
        #return warning and empty list for errors
        elif raises == 'warning':
            warning_list = [error_row, error_col, error_message, error_val]
        
        return error_list, warning_list        


    def generate_content_error(
        val_rule: str,
        attribute_name: str,
        sg: SchemaGenerator,
        row_num = None,
        error_val = None,    
    ) -> (List[str], List[str]):
        """
        Purpose:
            Generate an logging error or warning as well as a stored error/warning message when validating the content of a manifest attribute.

            Types of error/warning included:
                - recommended - Raised when an attribute is empty and recommended but not required.
                - unique - Raised when attribute values are not unique.
                - protectAges - Raised when an attribute contains ages below 18YO or over 90YO that should be censored.
        Input:
                val_rule: str, defined in the schema.
                attribute_name: str, attribute being validated
                sg: schemaGenerator object
                row_num: str, row where the error was detected
                error_val: value duplicated

        Returns:
            Logging.error or Logging.warning.
            Message: List[str] Error|Warning details for further storage.
        """
        error_list = []
        warning_list = []
        error_col = attribute_name  # Attribute name
        
        #Determine which, if any, message to raise
        raises = GenerateError.get_message_level(
            val_rule=val_rule,
            attribute_name = attribute_name,
            sg = sg,
            )

        #if a message needs to be raised, get the approrpiate function to do so
        if raises:
            logLevel = getattr(logging,raises)  
        else:
            return error_list, warning_list
        
        #log warning or error message
        if val_rule.startswith('recommended'):
            cross_error_str = (
                f"Column {attribute_name} is recommended but empty."
            )
            logLevel(cross_error_str)
            error_message = cross_error_str

            if raises == 'error':
                error_list = [error_col, error_message]
            #return warning and empty list for errors
            elif raises == 'warning':
                warning_list = [error_col, error_message]

            return error_list, warning_list

        elif val_rule.startswith('unique'):    
            cross_error_str = (
                f"Column {attribute_name} has the duplicate value(s) {set(error_val)} in rows: {row_num}."
            )

        elif val_rule.startswith('protectAges'):
            cross_error_str = (
                f"Column {attribute_name} contains ages that should be censored in rows: {row_num}."
            )           

        elif val_rule.startswith('inRange'):
            cross_error_str = (
                f"{attribute_name} values in rows {row_num} are out of the specified range."
            )  

        logLevel(cross_error_str)
        error_row = row_num 
        error_message = cross_error_str

        #return error and empty list for warnings
        if raises == 'error':
            error_list = [error_row, error_col, error_message, set(error_val)]
        #return warning and empty list for errors
        elif raises == 'warning':
            warning_list = [error_row, error_col, error_message, set(error_val)]
        
        return error_list, warning_list

    def get_message_level(
        sg: SchemaGenerator,
        attribute_name: str,
        val_rule: str,
        ) -> str:
        """
        Purpose:
            Determine whether an error or warning message should be logged and displayed
            
            if node is not required, 
                return warning
            if node is recommended and requried, 
                return None    
            for other rules, parse possible, if not use default specified in validation_rule_info

        Input:
                val_rule: str, defined in the schema.
                sg: schemaGenerator object
                attribute_name: str, attribute being validated
        Returns:
            'error', 'warning' or None
        # TODO: recommended and other rules
        """

        level = None
        rule_parts = val_rule.split(" ")
        rule_info = validation_rule_info()

        if not sg.is_node_required(node_display_name=attribute_name):
            # raise warning if recommended but not required
            if 'recommended' in val_rule:
                level = 'warning'
            # If not required or recommended raise warnings to notify
            else:
                level = 'warning' 
                return level
        elif sg.is_node_required(node_display_name=attribute_name) and 'recommended' in val_rule:
            level = None
        
        
        # Parse rule for level, set to default if not specified
        if rule_parts[-1].lower() == 'error' or rule_parts[0] == 'schema':
            level = 'error'
        elif rule_parts[-1].lower() == 'warning':
            level = 'warning'
        else:
            level = rule_info[rule_parts[0]]['default_message_level']
            
        return level

class ValidateAttribute(object):
    """
    A collection of functions to validate manifest attributes.
        list_validation
        regex_validation
        type_validation
        url_validation
        cross_validation
        get_target_manifests - helper function
    See functions for more details.
    TODO:
        - Add year validator
        - Add string length validator
    """

    def get_target_manifests(target_component, project_scope: List):

        target_manifest_IDs=[]
        target_dataset_IDs=[]
        
        #login
        access_token = getenv("SYNAPSE_ACCESS_TOKEN")
        if access_token:
            synStore = SynapseStorage(access_token=access_token,project_scope=project_scope)
        else:
            synStore = SynapseStorage(project_scope=project_scope)        

        #Get list of all projects user has access to
        projects = synStore.getStorageProjects(project_scope=project_scope)
        for project in projects:
            
            #get all manifests associated with datasets in the projects
            target_datasets=synStore.getProjectManifests(projectId=project[0])

            #If the manifest includes the target component, include synID in list
            for target_dataset in target_datasets:
                if target_component == target_dataset[-1][0].replace(" ","").lower() and target_dataset[1][0] != "":
                    target_manifest_IDs.append(target_dataset[1][0])
                    target_dataset_IDs.append(target_dataset[0][0])

        return synStore, target_manifest_IDs, target_dataset_IDs    

    def list_validation(
        self, val_rule: str, manifest_col: pd.core.series.Series, sg: SchemaGenerator,
    ) -> (List[List[str]], List[List[str]], pd.core.series.Series):
        """
        Purpose:
            Determine if values for a particular attribute are comma separated.
        Input:
            - val_rule: str, Validation rule
            - manifest_col: pd.core.series.Series, column for a given attribute
        Returns:
            - manifest_col: Input values in manifest arere-formatted to a list
            - Error log, error list
        """

        # For each 'list' (input as a string with a , delimiter) entered,
        # convert to a real list of strings, with leading and trailing
        # white spaces removed.
        errors = []
        warnings = []
        manifest_col = manifest_col.astype(str)
        csv_re = comma_separated_list_regex()

        rule_parts=val_rule.lower().split(" ")
        if len(rule_parts) > 1:
            list_robustness=rule_parts[1]
        else:
            list_robustness = 'strict'


        if list_robustness == 'strict':
        # This will capture any if an entry is not formatted properly. Only for strict lists
            for i, list_string in enumerate(manifest_col):
                if not re.fullmatch(csv_re,list_string):
                    list_error = "not_comma_delimited"
                    vr_errors, vr_warnings = GenerateError.generate_list_error(
                            list_string,
                            row_num=str(i + 2),
                            attribute_name=manifest_col.name,
                            list_error=list_error,
                            invalid_entry=manifest_col[i],
                            sg = sg,
                            val_rule = val_rule, 
                        )
                    if vr_errors:
                        errors.append(vr_errors)
                    if vr_warnings:
                        warnings.append(vr_warnings)
                

        # Convert string to list.
        manifest_col = parse_str_series_to_list(manifest_col)

        return errors, warnings, manifest_col

    def regex_validation(
        self, val_rule: str, manifest_col: pd.core.series.Series, sg: SchemaGenerator,
    ) -> (List[List[str]], List[List[str]]):
        """
        Purpose:
            Check if values for a given manifest attribue conform to the reguar expression,
            provided in val_rule.
        Input:
            - val_rule: str, Validation rule
            - manifest_col: pd.core.series.Series, column for a given
                attribute in the manifest
            Using this module requres validation rules written in the following manner:
                'regex module regular expression'
                - regex: is an exact string specifying that the input is to be validated as a 
                regular expression.
                - module: is the name of the module within re to run ie. search. 
                - regular_expression: is the regular expression with which to validate
                the user input.
        Returns:
            - This function will return errors when the user input value
            does not match schema specifications.
            Logging.error.
            Errors: List[str] Error details for further storage.
        TODO: 
            move validation to convert step.
        """

        reg_exp_rules = val_rule.split(" ")

        try:
            module_to_call = getattr(re, reg_exp_rules[1])
            reg_expression = reg_exp_rules[2]
        except:
            raise ValidationError(
                f"The regex rules were not provided properly for attribute {manifest_col.name}."
                f" They should be provided as follows ['regex', 'module name', 'regular expression']"
            )

        errors = []
        warnings = []
        validation_rules=self.sg.se.get_class_validation_rules(self.sg.se.get_class_label_from_display_name(manifest_col.name))
        # Handle case where validating re's within a list.
        if re.search('list',"|".join(validation_rules)):
            if type(manifest_col[0]) == str:
                # Convert string to list.
                manifest_col = parse_str_series_to_list(manifest_col)

            for i, row_values in enumerate(manifest_col):
                for j, re_to_check in enumerate(row_values):
                    re_to_check = str(re_to_check)
                    if not bool(module_to_call(reg_expression, re_to_check)) and bool(
                        re_to_check
                    ):
                        vr_errors, vr_warnings = GenerateError.generate_regex_error(
                                val_rule = val_rule,
                                reg_expression = reg_expression,
                                row_num=str(i + 2),
                                module_to_call=reg_exp_rules[1],
                                attribute_name=manifest_col.name,
                                invalid_entry=manifest_col[i],
                                sg = sg,
                            )
                        if vr_errors:
                            errors.append(vr_errors)
                        if vr_warnings:
                            warnings.append(vr_warnings)

        # Validating single re's
        else:
            manifest_col = manifest_col.astype(str)
            for i, re_to_check in enumerate(manifest_col):
                if not bool(module_to_call(reg_expression, re_to_check)) and bool(
                    re_to_check
                ):
                    vr_errors, vr_warnings = GenerateError.generate_regex_error(
                            val_rule = val_rule,
                            reg_expression = reg_expression,
                            row_num=str(i + 2),
                            module_to_call=reg_exp_rules[1],
                            attribute_name=manifest_col.name,
                            invalid_entry=manifest_col[i],
                            sg = sg,
                        )
                    if vr_errors:
                        errors.append(vr_errors)
                    if vr_warnings:
                        warnings.append(vr_warnings)

        return errors, warnings

    def type_validation(
        self, val_rule: str, manifest_col: pd.core.series.Series, sg: SchemaGenerator,
    ) -> (List[List[str]], List[List[str]]):
        """
        Purpose:
            Check if values for a given manifest attribue are the same type
            specified in val_rule.
        Input:
            - val_rule: str, Validation rule, specifying input type, either
                'float', 'int', 'num', 'str'
            - manifest_col: pd.core.series.Series, column for a given
                attribute in the manifest
        Returns:
            -This function will return errors when the user input value
            does not match schema specifications.
            Logging.error.
            Errors: List[str] Error details for further storage.
        TODO:
            Convert all inputs to .lower() just to prevent any entry errors.
        """
        specified_type = {
            'num': (int, np.int64, float),
            'int': (int, np.int64),
            'float': (float),
            'str': (str),
        }

        errors = []
        warnings = []
        # num indicates either a float or int.
        if val_rule == "num":
            for i, value in enumerate(manifest_col):
                if bool(value) and not isinstance(value, specified_type[val_rule]):
                    vr_errors, vr_warnings = GenerateError.generate_type_error(
                            val_rule = val_rule ,
                            row_num=str(i + 2),
                            attribute_name=manifest_col.name,
                            invalid_entry=str(manifest_col[i]),
                            sg = sg,
                        )
                    if vr_errors:
                        errors.append(vr_errors)
                    if vr_warnings:
                        warnings.append(vr_warnings)
        elif val_rule in ["int", "float", "str"]:
            for i, value in enumerate(manifest_col):
                if bool(value) and not isinstance(value, specified_type[val_rule]):
                    vr_errors, vr_warnings = GenerateError.generate_type_error(
                            val_rule = val_rule,
                            row_num=str(i + 2),
                            attribute_name=manifest_col.name,
                            invalid_entry=str(manifest_col[i]),
                            sg = sg,
                        )
                    if vr_errors:
                        errors.append(vr_errors)
                    if vr_warnings:
                        warnings.append(vr_warnings)
        return errors, warnings

    def url_validation(self, val_rule: str, manifest_col: str, sg: SchemaGenerator,) -> (List[List[str]], List[List[str]]):
        """
        Purpose:
            Validate URL's submitted for a particular attribute in a manifest.
            Determine if the URL is valid and contains attributes specified in the
            schema.
        Input:
            - val_rule: str, Validation rule
            - manifest_col: pd.core.series.Series, column for a given
                attribute in the manifest
        Output:
            This function will return errors when the user input value
            does not match schema specifications.
        """

        url_args = val_rule.split(" ")[1:]
        errors = []
        warnings = []

        for i, url in enumerate(manifest_col):
            # Check if a random phrase, string or number was added and
            # log the appropriate error.
            if not isinstance(url,str) or not (
                urlparse(url).scheme
                + urlparse(url).netloc
                + urlparse(url).params
                + urlparse(url).query
                + urlparse(url).fragment
            ):
                #
                url_error = "random_entry"
                valid_url = False
                vr_errors, vr_warnings = GenerateError.generate_url_error(
                        url,
                        url_error=url_error,
                        row_num=str(i + 2),
                        attribute_name=manifest_col.name,
                        argument=url_args,
                        invalid_entry=manifest_col[i],
                        sg = sg,
                        val_rule = val_rule,
                    )
                if vr_errors:
                    errors.append(vr_errors)
                if vr_warnings:
                    warnings.append(vr_warnings)
            else:
                # add scheme to the URL if not currently added.
                if not urlparse(url).scheme:
                    url = "http://" + url
                try:
                    # Check that the URL points to a working webpage
                    # if not log the appropriate error.
                    request = Request(url)
                    response = urlopen(request)
                    valid_url = True
                    response_code = response.getcode()
                except:
                    valid_url = False
                    url_error = "invalid_url"
                    vr_errors, vr_warnings = GenerateError.generate_url_error(
                            url,
                            url_error=url_error,
                            row_num=str(i + 2),
                            attribute_name=manifest_col.name,
                            argument=url_args,
                            invalid_entry=manifest_col[i],
                            sg = sg,
                            val_rule = val_rule,
                        )
                    if vr_errors:
                        errors.append(vr_errors)
                    if vr_warnings:
                        warnings.append(vr_warnings)
                if valid_url == True:
                    # If the URL works, check to see if it contains the proper arguments
                    # as specified in the schema.
                    for arg in url_args:
                        if arg not in url:
                            url_error = "arg_error"
                            vr_errors, vr_warnings = GenerateError.generate_url_error(
                                    url,
                                    url_error=url_error,
                                    row_num=str(i + 2),
                                    attribute_name=manifest_col.name,
                                    argument=arg,
                                    invalid_entry=manifest_col[i],
                                    sg = sg,
                                    val_rule = val_rule,
                                )
                            if vr_errors:
                                errors.append(vr_errors)
                            if vr_warnings:
                                warnings.append(vr_warnings)
        return errors, warnings

    def cross_validation(
        self, val_rule: str, manifest_col: pd.core.series.Series, project_scope: List, sg: SchemaGenerator,
    ) -> List[List[str]]:
        """
        Purpose:
            Do cross validation between the current manifest and all other manifests a user has access to on Synapse.
            Check if values in this manifest are present fully in others.
        Input:
            - val_rule: str, Validation rule
            - manifest_col: pd.core.series.Series, column for a given
                attribute in the manifest
        Output:
            This function will return errors when values in the current manifest's attribute 
            are not fully present in the correct amount of other manifests.
        """
        errors = []
        warnings = []
        missing_values = {}
        missing_manifest_log={}
        present_manifest_log=[]
        target_column = pd.Series(dtype=object)
        #parse sources and targets
        source_attribute=manifest_col.name
        [target_component, target_attribute] = val_rule.lower().split(" ")[1].split(".")
        scope=val_rule.lower().split(" ")[2]
        target_column.name=target_attribute

        
        #Get IDs of manifests with target component
        synStore, target_manifest_IDs, target_dataset_IDs = ValidateAttribute.get_target_manifests(target_component,project_scope)

        #Read each manifest
        for target_manifest_ID, target_dataset_ID in zip(target_manifest_IDs,target_dataset_IDs):
            entity = synStore.getDatasetManifest(
                datasetId = target_dataset_ID,
                downloadFile = True
                )
            target_manifest=pd.read_csv(entity.path)

            #convert manifest column names into validation rule input format - 
            column_names={}
            for name in target_manifest.columns:
                column_names[name.replace(" ","").lower()]=name

            if scope.__contains__('set'):
                #If the manifest has the target attribute for the component do the cross validation
                if target_attribute in column_names:                    
                    target_column = target_manifest[column_names[target_attribute]]

                    #Do the validation on both columns
                    missing_values = manifest_col[~manifest_col.isin(target_column)]

                    if missing_values.empty:
                        present_manifest_log.append(target_manifest_ID)
                    else:
                        missing_manifest_log[target_manifest_ID] = missing_values

            elif scope.__contains__('value'):
                if target_attribute in column_names:
                    target_manifest.rename(columns={column_names[target_attribute]: target_attribute}, inplace=True)
                    
                    target_column = pd.concat(
                        objs = [target_column, target_manifest[target_attribute]],
                        join = 'outer',
                        ignore_index= True,
                    )                
                    target_column = target_column.astype('object')
                    #print(target_column)
                    
        
        
        missing_rows=[]
        missing_values=[]  


        if scope.__contains__('value'):
            missing_values = manifest_col[~manifest_col.isin(target_column)]
            duplicated_values = manifest_col[manifest_col.isin(target_column[target_column.duplicated()])]
            
            if val_rule.__contains__('matchAtLeastOne') and not missing_values.empty:
                missing_rows = missing_values.index.to_numpy() + 2
                vr_errors, vr_warnings = GenerateError.generate_cross_warning(
                        val_rule = val_rule,
                        row_num = str(list(missing_rows)),
                        attribute_name = source_attribute,
                        invalid_entry = str(missing_values.values.tolist()),
                        sg = sg,
                    )
                if vr_errors:
                    errors.append(vr_errors)
                if vr_warnings:
                    warnings.append(vr_warnings)
            elif val_rule.__contains__('matchExactlyOne') and (duplicated_values.any() or missing_values.any()):
                invalid_values  = pd.merge(duplicated_values,missing_values,how='outer')
                invalid_rows    = pd.merge(duplicated_values,missing_values,how='outer',left_index=True,right_index=True).index.to_numpy() + 2
                vr_errors, vr_warnings = GenerateError.generate_cross_warning(
                        val_rule = val_rule,
                        row_num = str(list(invalid_rows)), 
                        attribute_name = source_attribute, 
                        invalid_entry = str(pd.Series(invalid_values.squeeze()).values.tolist()),
                        sg = sg,
                    )
                if vr_errors:
                    errors.append(vr_errors)
                if vr_warnings:
                    warnings.append(vr_warnings)
            

            
        #generate warnings if necessary
        elif scope.__contains__('set'):
            if val_rule.__contains__('matchAtLeastOne') and len(present_manifest_log) < 1:     
                missing_entries = list(missing_manifest_log.values()) 
                missing_manifest_IDs = list(missing_manifest_log.keys()) 
                for missing_entry in missing_entries:
                    missing_rows.append(missing_entry.index[0]+2)
                    missing_values.append(missing_entry.values[0])
                    
                missing_rows=list(set(missing_rows))
                missing_values=list(set(missing_values))
                #print(missing_rows,missing_values)

                vr_errors, vr_warnings = GenerateError.generate_cross_warning(
                        val_rule = val_rule,
                        row_num = str(missing_rows),
                        attribute_name = source_attribute,
                        invalid_entry = str(missing_values),
                        missing_manifest_ID = missing_manifest_IDs,
                        sg = sg,
                    )
                if vr_errors:
                    errors.append(vr_errors)
                if vr_warnings:
                    warnings.append(vr_warnings)
            elif val_rule.__contains__('matchExactlyOne') and len(present_manifest_log) != 1:
                vr_errors, vr_warnings = GenerateError.generate_cross_warning(
                        val_rule = val_rule,
                        attribute_name = source_attribute,
                        matching_manifests = present_manifest_log,
                        sg = sg,
                    )
                if vr_errors:
                    errors.append(vr_errors)
                if vr_warnings:
                    warnings.append(vr_warnings)
                

        return errors, warnings





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['schematic',
 'schematic.manifest',
 'schematic.models',
 'schematic.schemas',
 'schematic.store',
 'schematic.utils',
 'schematic.visualization']

package_data = \
{'': ['*'],
 'schematic': ['etc/*', 'etc/data_models/*', 'etc/validation_schemas/*']}

install_requires = \
['Flask-Cors>=3.0.10,<4.0.0',
 'Flask>=1.1.4,<2.0.0',
 'Jinja2==2.11.3',
 'MarkupSafe==2.0.1',
 'PyYAML>=5.4.1,<6.0.0',
 'click-log>=0.3.2,<0.4.0',
 'click>=7.1.2,<8.0.0',
 'connexion[swagger-ui]>=2.8.0,<3.0.0',
 'google-api-python-client>=1.12.8,<2.0.0',
 'google-auth-httplib2>=0.0.4,<0.0.5',
 'google-auth-oauthlib>=0.4.2,<0.5.0',
 'graphviz>=0.16,<0.17',
 'great-expectations>=0.15.0,<0.16.0',
 'inflection>=0.5.1,<0.6.0',
 'itsdangerous==1.1.0',
 'jsonschema>=3.2.0,<4.0.0',
 'networkx>=2.5,<3.0',
 'numpy>=1.21.1,<2.0.0',
 'oauth2client<4.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pandas>=1.3.1,<2.0.0',
 'pdoc>=12.2.0,<13.0.0',
 'pygsheets>=2.0.4,<3.0.0',
 'rdflib>=5.0.0,<6.0.0',
 'setuptools>=52.0.0,<53.0.0',
 'sphinx-click>=3.1.0,<4.0.0',
 'synapseclient>=2.6.0,<3.0.0',
 'tenacity>=8.0.1,<9.0.0',
 'toml>=0.10.2,<0.11.0']

extras_require = \
{':python_version < "3.9"': ['backports.zoneinfo>=0.2.1,<0.3.0']}

entry_points = \
{'console_scripts': ['schematic = schematic.__main__:main']}

setup_kwargs = {
    'name': 'schematicpy',
    'version': '22.11.3',
    'description': 'Package for biomedical data model and metadata ingress management',
    'long_description': '# Schematic\n[![Build Status](https://img.shields.io/endpoint.svg?url=https%3A%2F%2Factions-badge.atrox.dev%2FSage-Bionetworks%2Fschematic%2Fbadge%3Fref%3Ddevelop&style=flat)](https://actions-badge.atrox.dev/Sage-Bionetworks/schematic/goto?ref=develop) [![Documentation Status](https://readthedocs.org/projects/sage-schematic/badge/?version=develop)](https://sage-schematic.readthedocs.io/en/develop/?badge=develop) [![PyPI version](https://badge.fury.io/py/schematicpy.svg)](https://badge.fury.io/py/schematicpy)\n\n# Table of contents\n- [Introduction](#introduction)\n- [Installation](#installation)\n  - [Installation Requirements](#installation-requirements)\n  - [Installation guide for data curator app](#installation-guide-for-data-curator-app)\n  - [Installation guide for developers/contributors](#installation-guide-for-developerscontributors)\n- [Other Contribution Guidelines](#other-contribution-guidelines)\n    - [Update readthedocs documentation](#update-readthedocs-documentation)\n- [Command Line Usage](#command-line-usage)\n- [Testing](#testing)\n  - [Updating Synapse test resources](#updating-synapse-test-resources)\n- [Code Style](#code-style)\n- [Contributors](#contributors)\n\n# Introduction\nSCHEMATIC is an acronym for _Schema Engine for Manifest Ingress and Curation_. The Python based infrastructure provides a _novel_ schema-based, metadata ingress ecosystem, that is meant to streamline the process of biomedical dataset annotation, metadata validation and submission to a data repository for various data contributors.\n\n# Installation\n## Installation Requirements\n* Python 3.7.1 or higher\n\nNote: You need to be a registered and certified user on [`synapse.org`](https://www.synapse.org/), and also have the right permissions to download the Google credentials files from Synapse.\n\n\n## Installation guide for data curator app\n\nCreate and activate a virtual environment within which you can install the package:\n\n```\npython3 -m venv .venv\nsource .venv/bin/activate\n```\n\nNote: Python 3 has a built-in support for virtual environment [venv](https://docs.python.org/3/library/venv.html#module-venv) so you no longer need to install virtualenv.\n\nInstall and update the package using [pip](https://pip.pypa.io/en/stable/quickstart/):\n\n```\npython3 -m pip install schematicpy\n```\n\nIf you run into error: Failed building wheel for numpy, the error might be able to resolve by upgrading pip. Please try to upgrade pip by:\n\n```\npip3 install --upgrade pip\n```\n\n## Installation guide for developers/contributors \n\nWhen contributing to this repository, please first discuss the change you wish to make via issue, email, or any other method with the owners of this repository before making a change.\n\nPlease note we have a [code of conduct](CODE_OF_CONDUCT.md), please follow it in all your interactions with the project.\n\n### Development environment setup\n1. Clone the `schematic` package repository.\n```\ngit clone https://github.com/Sage-Bionetworks/schematic.git\n```\n2. Install `poetry` (version 1.2 or later) using either the [official installer](https://python-poetry.org/docs/#installing-with-the-official-installer) or [pipx](https://python-poetry.org/docs/#installing-with-pipx). If you have an older installation of Poetry, we recommend uninstalling it first. \n\n3. Start the virtual environment by doing: \n```\npoetry shell\n```\n4. Install the dependencies by doing: \n```\npoetry install\n```\nThis command will install the dependencies based on what we specify in poetry.lock. If this step is taking a long time, try to go back to step 2 and check your version of poetry. Alternatively, you could also try deleting the lock file and regenerate it by doing `poetry install` (Please note this method should be used as a last resort because this would force other developers to change their development environment)\n\n5. Fill in credential files: \n*Note*: If you won\'t interact with Synapse, please ignore this section.\n\nThere are two main configuration files that need to be edited:\n[config.yml](https://github.com/Sage-Bionetworks/schematic/blob/develop/config.yml)\nand [synapseConfig](https://raw.githubusercontent.com/Sage-Bionetworks/synapsePythonClient/v2.3.0-rc/synapseclient/.synapseConfig)\n\n<strong>Configure .synapseConfig File</strong>\n\nDownload a copy of the ``.synapseConfig`` file, open the file in the\neditor of your choice and edit the `username` and `authtoken` attribute under the `authentication` section \n\n*Note*: You could also visit [configparser](https://docs.python.org/3/library/configparser.html#module-configparser>) doc to see the format that `.synapseConfig` must have. For instance:\n>[authentication]<br> username = ABC <br> authtoken = abc\n\n<strong>Configure config.yml File</strong>\n\n*Note*: Below is only a brief explanation of some attributes in `config.yml`. <strong>Please use the link [here](https://github.com/Sage-Bionetworks/schematic/blob/develop/config.yml) to get the latest version of `config.yml` in `develop` branch</strong>.\n\nDescription of `config.yml` attributes\n\n    definitions:\n        synapse_config: "~/path/to/.synapseConfig"\n        creds_path: "~/path/to/credentials.json"\n        token_pickle: "~/path/to/token.pickle"\n        service_acct_creds: "~/path/to/service_account_creds.json"\n\n    synapse:\n        master_fileview: "syn23643253" # fileview of project with datasets on Synapse\n        manifest_folder: "~/path/to/manifest_folder/" # manifests will be downloaded to this folder\n        manifest_basename: "filename" # base name of the manifest file in the project dataset, without extension\n        token_creds: "syn23643259" # synapse ID of credentials.json file\n        service_acct_creds: "syn25171627" # synapse ID of service_account_creds.json file\n\n    manifest:\n        title: "example" # title of metadata manifest file\n        # to make all manifests enter only \'all manifests\'\n        data_type: \n          - "Biospecimen"\n          - "Patient"\n\n    model:\n        input:\n            location: "data/schema_org_schemas/example.jsonld" # path to JSON-LD data model\n            file_type: "local" # only type "local" is supported currently\n    style: # configuration of google sheet\n        google_manifest:\n          req_bg_color:\n            red: 0.9215\n            green: 0.9725\n            blue: 0.9803\n          opt_bg_color:\n            red: 1.0\n            green: 1.0\n            blue: 0.9019\n          master_template_id: \'1LYS5qE4nV9jzcYw5sXwCza25slDfRA1CIg3cs-hCdpU\'\n          strict_validation: true\n\n*Note*: Paths can be specified relative to the `config.yml` file or as absolute paths.\n\n6. Login to Synapse by using the command line\nOn the CLI in your virtual environment, run the following command: \n```\nsynapse login -u <synapse username> -p <synapse password> --rememberMe\n```\nPlease make sure that you run the command before running `schematic init` below\n\n7. Obtain Google credential Files\n\nTo obtain ``credentials.json`` and ``token.pickle``, please run:\n\n```\nschematic init --config ~/path/to/config.yml\n```\nThis should prompt you with a URL that will take you through Google OAuth. Your `credentials.json` and `token.pickle` will get automatically downloaded the first time you run this command.\n\n*Note* : The ``credentials.json`` file is required when you are using\n[OAuth2](https://developers.google.com/identity/protocols/oauth2)\nto authenticate with the Google APIs.\n\nFor details about the steps involved in the [OAuth2 authorization\nflow](https://github.com/Sage-Bionetworks/schematic/blob/develop/schematic/utils/google_api_utils.py#L18)\nrefer to the ``Credentials`` section in the\n[docs/md/details](https://github.com/Sage-Bionetworks/schematic/blob/develop/docs/md/details.md#credentials)\ndocument.\n\nTo obtain  ``schematic_service_account_creds.json``, please run: \n```\nschematic init --config ~/path/to/config.yml --auth service_account\n```\n*Notes*: Use the ``schematic_service_account_creds.json`` file for the service\naccount mode of authentication (*for Google services/APIs*). Service accounts \nare special Google accounts that can be used by applications to access Google APIs \nprogrammatically via OAuth2.0, with the advantage being that they do not require \nhuman authorization. \n\n*Background*: schematic uses Google’s API to generate google sheet templates that users fill in to provide (meta)data.\nMost Google sheet functionality could be authenticated with service account. However, more complex Google sheet functionality\nrequires token-based authentication. As browser support that requires the token-based authentication diminishes, we are hoping to deprecate\ntoken-based authentication and keep only service account authentication in the future. \n\n\n### Development process instruction\n\nFor new features, bugs, enhancements\n\n1. Pull the latest code from [develop branch in the upstream repo](https://github.com/Sage-Bionetworks/schematic)\n2. Checkout a new branch develop-<feature/fix-name> from the develop branch\n3. Do development on branch develop-<feature/fix-name>\n   a. may need to ensure that schematic poetry toml and lock files are compatible with your local environment\n4. Add changed files for tracking and commit changes using [best practices](https://www.perforce.com/blog/vcs/git-best-practices-git-commit)\n5. Have granular commits: not “too many” file changes, and not hundreds of code lines of changes\n6. Commits with work in progress are encouraged:\n   a. add WIP to the beginning of the commit message for “Work In Progress” commits\n7. Keep commit messages descriptive but less than a page long, see best practices\n8. Push code to develop-<feature/fix-name> in upstream repo\n9. Branch out off develop-<feature/fix-name> if needed to work on multiple features associated with the same code base\n10. After feature work is complete and before creating a PR to the develop branch in upstream\n    a. ensure that code runs locally\n    b. test for logical correctness locally\n    c. wait for git workflow to complete (e.g. tests are run) on github\n11. Create a PR from develop-<feature/fix-name> into the develop branch of the upstream repo\n12. Request a code review on the PR\n13. Once code is approved merge in the develop branch\n14. Delete the develop-<feature/fix-name> branch\n\n*Note*: Make sure you have the latest version of the `develop` branch on your local machine.\n\n## Installation Guide - Docker \n\n1. Install docker from https://www.docker.com/ . <br>\n2.  Identify docker image of interest from [Schematic DockerHub](https://hub.docker.com/r/sagebionetworks/schematic/tags) <br>\n    Ex `docker pull sagebionetworks/schematic:latest` from the CLI or, run `docker compose up` after cloning the schematic github repo <br>\n    in this case, `sagebionetworks/schematic:latest` is the name of the image chosen\n3. Run Schematic Command with `docker run <flags> <schematic command and args>`. <br>\n<t> - For more information on flags for `docker run` and what they do, visit the [Docker Documentation](https://docs.docker.com/engine/reference/commandline/run/) <br>\n<t> - These example commands assume that you have navigated to the directory you want to run schematic from. To specify your working directory, use `$(pwd)` on MacOS/Linux or `%cd%` on Windows.  <br>\n<t> - If not using the latest image, then the full name should be specified: ie `sagebionetworks/schematic:commit-e611e4a` <br>\n<t> - If using local image created by `docker compose up`, then the docker image name should be changed: i.e. `schematic_schematic` <br>\n<t> - Using the `--name` flag sets the name of the container running locally on your machine <br>\n\n### Example For REST API <br>\n\n#### Use file path of `config.yml` to run API endpoints: \n```\ndocker run --rm -p 3001:3001 \\\n  -v $(pwd):/schematic -w /schematic --name schematic \\\n  -e SCHEMATIC_CONFIG=/schematic/config.yml \\\n  -e GE_HOME=/usr/src/app/great_expectations/ \\\n  sagebionetworks/schematic \\\n  python /usr/src/app/run_api.py\n``` \n\n#### Use content of `config.yml` as an environment variable to run API endpoints: \n1. save content of `config.yml` as to environment variable `SCHEMATIC_CONFIG_CONTENT` by doing: `export SCHEMATIC_CONFIG_CONTENT=$(cat config.yml)`\n\n2. Pass `SCHEMATIC_CONFIG_CONTENT` as an environment variable by using `docker run`\n\n```\ndocker run --rm -p 3001:3001 \\\n  -v $(pwd):/schematic -w /schematic --name schematic \\\n  -e GE_HOME=/usr/src/app/great_expectations/ \\\n  -e SCHEMATIC_CONFIG_CONTENT=$SCHEMATIC_CONFIG_CONTENT \\\n  sagebionetworks/schematic \\\n  python /usr/src/app/run_api.py\n``` \n\n\n### Example For Schematic on mac/linux <br>\nTo run example below, first clone schematic into your home directory  `git clone https://github.com/sage-bionetworks/schematic ~/schematic` <br>\nThen update .synapseConfig with your credentials\n```\ndocker run \\\n  -v ~/schematic:/schematic \\\n  -w /schematic \\\n  -e SCHEMATIC_CONFIG=/schematic/config.yml \\\n  -e GE_HOME=/usr/src/app/great_expectations/ \\\n  sagebionetworks/schematic schematic model \\\n  -c /schematic/config.yml validate \\\n  -mp /schematic/tests/data/mock_manifests/Valid_Test_Manifest.csv \\\n  -dt MockComponent \\\n  -js /schematic/tests/data/example.model.jsonld\n``` \n\n### Example For Schematic on Windows <br>\n```\ndocker run -v %cd%:/schematic \\\n  -w /schematic \\\n  -e GE_HOME=/usr/src/app/great_expectations/ \\\n  sagebionetworks/schematic \\\n  schematic model \\\n  -c config.yml validate -mp tests/data/mock_manifests/inValid_Test_Manifest.csv -dt MockComponent -js /schematic/data/example.model.jsonld\n```\n\n# Other Contribution Guidelines\n## Updating readthedocs documentation\n1. `cd docs`\n2. After making relevant changes, you could run the `make html` command to re-generate the `build` folder.\n3. Please contact the dev team to publish your updates\n\n*Other helpful resources*:\n\n1. [Getting started with Sphinx](https://haha.readthedocs.io/en/latest/intro/getting-started-with-sphinx.html)\n2. [Installing Sphinx](https://haha.readthedocs.io/en/latest/intro/getting-started-with-sphinx.html)\n\n## Update toml file and lock file\nIf you install external libraries by using `poetry add <name of library>`, please make sure that you include `pyproject.toml` and `poetry.lock` file in your commit.\n\n## Reporting bugs or feature requests\nYou can use the [`Issues`](https://github.com/Sage-Bionetworks/schematic/issues) tab to **create bug and feature requests**. Providing enough details to the developers to verify and troubleshoot your issue is paramount:\n- **Provide a clear and descriptive title as well as a concise summary** of the issue to identify the problem.\n- **Describe the exact steps which reproduce the problem** in as many details as possible.\n- **Describe the behavior you observed after following the steps** and point out what exactly is the problem with that behavior.\n- **Explain which behavior you expected to see** instead and why.\n- **Provide screenshots of the expected or actual behaviour** where applicable.\n\n# Command Line Usage\nPlease visit more documentation [here](https://sage-schematic.readthedocs.io/en/develop/cli_reference.html)\n\n\n\n# Testing \n\nAll code added to the client must have tests. The Python client uses pytest to run tests. The test code is located in the [tests](https://github.com/Sage-Bionetworks/schematic/tree/develop-docs-update/tests) subdirectory.\n\nYou can run the test suite in the following way:\n\n```\npytest -vs tests/\n```\n\n## Updating Synapse test resources\n\n1. Duplicate the entity being updated (or folder if applicable).\n2. Edit the duplicates (_e.g._ annotations, contents, name).\n3. Update the test suite in your branch to use these duplicates, including the expected values in the test assertions.\n4. Open a PR as per the usual process (see above).\n5. Once the PR is merged, leave the original copies on Synapse to maintain support for feature branches that were forked from `develop` before your update.\n   - If the old copies are problematic and need to be removed immediately (_e.g._ contain sensitive data), proceed with the deletion and alert the other contributors that they need to merge the latest `develop` branch into their feature branches for their tests to work.\n\n# Code style\n\n* Please consult the [Google Python style guide](http://google.github.io/styleguide/pyguide.html) prior to contributing code to this project.\n* Be consistent and follow existing code conventions and spirit.\n\n\n# Contributors\n\nMain contributors and developers:\n\n- [Milen Nikolov](https://github.com/milen-sage)\n- [Mialy DeFelice](https://github.com/mialy-defelice)\n- [Sujay Patil](https://github.com/sujaypatil96)\n- [Bruno Grande](https://github.com/BrunoGrandePhD)\n- [Robert Allaway](https://github.com/allaway)\n- [Gianna Jordan](https://github.com/giajordan)\n- [Lingling Peng](https://github.com/linglp)\n',
    'author': 'Milen Nikolov',
    'author_email': 'milen.nikolov@sagebase.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Sage-Bionetworks/schematic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



import logging


logger = logging.getLogger(__name__)


def extract_name_from_uri_or_curie(item):
    """Extract name from uri or curie"""
    if "http" not in item and len(item.split(":")) == 2:
        return item.split(":")[-1]
    elif len(item.split("//")[-1].split("/")) > 1:
        return item.split("//")[-1].split("/")[-1]
    else:
        raise ValueError("Error extracting name from URI or Curie.")


def expand_curie_to_uri(curie, context_info):
    """Expand curie to uri based on the context given

    parmas
    ======
    curie: curie to be expanded (e.g. bts:BiologicalEntity)
    context_info: jsonld context specifying prefix-uri relation (e.g. {"bts":
    "http://schema.biothings.io/"})
    """
    # as suggested in SchemaOrg standard file, these prefixes don't expand
    PREFIXES_NOT_EXPAND = ["rdf", "rdfs", "xsd"]
    # determine if a value is curie
    if len(curie.split(":")) == 2:
        prefix, value = curie.split(":")
        if prefix in context_info and prefix not in PREFIXES_NOT_EXPAND:
            return context_info[prefix] + value
        # if the input is not curie, return the input unmodified
        else:
            return curie
    else:
        return curie


def expand_curies_in_schema(schema):
    """Expand all curies in a SchemaOrg JSON-LD file into URI"""
    context = schema["@context"]
    graph = schema["@graph"]
    new_schema = {"@context": context, "@graph": [], "@id": schema["@id"]}
    for record in graph:
        new_record = {}
        for k, v in record.items():
            if type(v) == str:
                new_record[expand_curie_to_uri(k, context)] = expand_curie_to_uri(
                    v, context
                )
            elif type(v) == list:
                if type(v[0]) == dict:
                    new_record[expand_curie_to_uri(k, context)] = []
                    for _item in v:
                        new_record[expand_curie_to_uri(k, context)].append(
                            {"@id": expand_curie_to_uri(_item["@id"], context)}
                        )
                else:
                    new_record[expand_curie_to_uri(k, context)] = [
                        expand_curie_to_uri(_item, context) for _item in v
                    ]
            elif type(v) == dict and "@id" in v:
                new_record[expand_curie_to_uri(k, context)] = {
                    "@id": expand_curie_to_uri(v["@id"], context)
                }
            elif v == None:
                new_record[expand_curie_to_uri(k, context)] = None
        new_schema["@graph"].append(new_record)
    return new_schema


def uri2label(uri, schema):
    """Given a URI, return the label"""
    return [
        record["rdfs:label"] for record in schema["@graph"] if record["@id"] == uri
    ][0]



"""CURIE - A Module for handling Compact URIs (CURIEs)

Features:
- Support for RDFLib datatypes (BNode, URIRef), but doesn't require it
- MIT licensed

CURIE Syntax specification:
    <http://www.w3.org/TR/2007/WD-curie-20070307/>
Blank node handling notes:
    <http://milicicvuk.com/blog/2011/08/18/extended-curie-prefixlocalnamekey/>

Copyright (C) 2011-2012 Jaakko Salonen

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

__author__ = "Jaakko Salonen"
__copyright__ = "Copyright 2011-2012, Jaakko Salonen"
__version__ = "0.5.0"
__license__ = "MIT"
__status__ = "Prototype"

from urllib.parse import unquote
from copy import copy
from rdflib import BNode, URIRef

import sys

if sys.version_info.major == 3:
    unicode = str


try:
    from rdflib import BNode, URIRef
except:
    # Fallback if rdflib is not present
    class BNode(object):
        def __init__(self, val):
            self.val = val

        def n3(self):
            return unicode("_:" + self.val)

    class URIRef(unicode):
        pass


class Curie(object):
    """Curie Datatype Class

    Examples:

    >>> nss = dict(dc='http://purl.org/dc/elements/1.1/')

    >>> dc_title = Curie('http://purl.org/dc/elements/1.1/title', nss)

    >>> dc_title.curie
    u'dc:title'

    >>> dc_title.uri
    u'http://purl.org/dc/elements/1.1/title'

    >>> dc_title.curie
    u'dc:title'

    >>> nss['example'] = 'http://www.example.org/'

    >>> iri_test = Curie('http://www.example.org/D%C3%BCrst', nss)

    >>> iri_test.uri
    u'http://www.example.org/D\\xfcrst'

    >>> iri_test.curie
    u'example:D%C3%BCrst'

    """

    def __init__(self, uri, namespaces=dict()):
        self.namespaces = namespaces
        self.uri = unicode(unquote(uri), "utf-8")
        self.curie = copy(self.uri)
        for ns in self.namespaces:
            self.curie = uri.replace("" + self.namespaces["%s" % ns], "%s:" % ns)

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.curie


def uri2curie(uri, namespaces):
    """Convert URI to CURIE

    Define namespaces we want to use:
    >>> nss = dict(dc='http://purl.org/dc/elements/1.1/')

    Converting a string URI to CURIE
    >>> uri2curie('http://purl.org/dc/elements/1.1/title', nss)
    u'dc:title'

    RDFLib data type conversions:

    URIRef to CURIE
    >>> uri2curie(URIRef('http://purl.org/dc/elements/1.1/title'), nss)
    u'dc:title'

    Blank node to CURIE
    >>> uri2curie(BNode('blanknode1'), nss)
    u'_:blanknode1'

    """

    # Use n3() method if BNode
    if isinstance(uri, BNode):
        result = uri.n3()
    else:
        result = uri

    #  result = unicode(uri)
    for ns in namespaces:
        ns_raw = "%s" % namespaces["%s" % ns]
        if ns_raw == "http://www.w3.org/2002/07/owl#uri":
            ns_raw = "http://www.w3.org/2002/07/owl#"
        result = result.replace(ns_raw, "%s:" % ns)

    result = result.replace("http://www.w3.org/2002/07/owl#", "owl:")
    return result


def curie2uri(curie, namespaces):
    """Convert CURIE to URI

    TODO: testing

    """
    result = unicode(curie)
    for ns in namespaces:
        result = result.replace("%s:" % ns, "" + namespaces["%s" % ns])
    return URIRef(result)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import os
import pandas as pd
from jsonschema import validate
from re import compile, search, IGNORECASE
from schematic.utils.io_utils import load_json
from schematic import CONFIG, LOADER
from typing import List

def validate_schema(schema):
    """Validate schema against schema.org standard"""
    data_path = "validation_schemas/model.schema.json"
    json_schema_path = LOADER.filename(data_path)
    json_schema = load_json(json_schema_path)
    return validate(schema, json_schema)


def validate_property_schema(schema):
    """Validate schema against SchemaORG property definition standard"""
    data_path = "validation_schemas/property.schema.json"
    json_schema_path = LOADER.filename(data_path)
    json_schema = load_json(json_schema_path)
    return validate(schema, json_schema)


def validate_class_schema(schema):
    """Validate schema against SchemaORG class definition standard"""
    data_path = "validation_schemas/class.schema.json"
    json_schema_path = LOADER.filename(data_path)
    json_schema = load_json(json_schema_path)
    return validate(schema, json_schema)

def comma_separated_list_regex():
    # Regex to match with comma separated list 
    # Requires at least one element and a comma to be valid 
    # Does not require a trailing comma
    csv_list_regex=compile('([^\,]+\,)(([^\,]+\,?)*)')

    return csv_list_regex

def rule_in_rule_list(rule: str, rule_list: List[str]):
    # Function to standardize 
    # checking to see if a rule is contained in a list of rules. 
    # Uses regex to avoid issues arising from validation rules with arguments 
    # or rules that have arguments updated.

    # seperate rule type if arguments are specified
    rule_type = rule.split(" ")[0]

    # Process string and list of strings for regex comparison
    rule_type = rule_type + '[^\|]*'
    rule_list = '|'.join(rule_list)

    return search(rule_type, rule_list, flags=IGNORECASE)

def parse_str_series_to_list(col: pd.Series):
    """
    Parse a pandas series of comma delimited strings
    into a series with values that are lists of strings 
    ex. 
        Input:  'a,b,c'
        Output: ['a','b','c']     

    """
    col = col.apply(
        lambda x: [s.strip() for s in str(x).split(",")]
    )

    return col


import os
import json
import urllib.request

from schematic import CONFIG, LOADER


def load_json(file_path):
    """Load json document from file path or url

    :arg str file_path: The path of the url doc, could be url or file path
    """
    if file_path.startswith("http"):
        with urllib.request.urlopen(file_path) as url:
            data = json.loads(url.read().decode())
            return data
    # handle file path
    else:
        with open(file_path, encoding="utf8") as f:
            data = json.load(f)
            return data


def export_json(json_doc, file_path):
    """Export JSON doc to file"""
    with open(file_path, "w", encoding="utf8") as f:
        json.dump(json_doc, f, sort_keys=True, indent=4, ensure_ascii=False)


def load_default():
    """Load biolink vocabulary"""
    data_path = "data_models/biothings.model.jsonld"
    biothings_path = LOADER.filename(data_path)

    return load_json(biothings_path)


def load_schemaorg():
    """Load SchemOrg vocabulary"""
    data_path = "data_models/schema_org.model.jsonld"
    schemaorg_path = LOADER.filename(data_path)

    return load_json(schemaorg_path)
